(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory DpTrans imports Innermost begin

types ('f, 'v) dpp = "('f, 'v) trs \<times> ('f, 'v) qtrs"

definition minimal_cond :: "('f,'v)qtrs \<Rightarrow> (nat \<Rightarrow> ('f,'v)term) \<Rightarrow> (nat \<Rightarrow> ('f,'v)term) \<Rightarrow> (nat \<Rightarrow> ('f,'v)subst) \<Rightarrow> bool"
where "minimal_cond QR s t \<sigma> \<equiv> \<forall> i. (SN_elt (qrstep QR) (t i \<cdot> \<sigma> i))"

fun ichain ::
 "('f,'v)dpp \<Rightarrow> (nat \<Rightarrow> ('f,'v)term) \<Rightarrow> (nat \<Rightarrow> ('f,'v)term) \<Rightarrow> (nat \<Rightarrow> ('f,'v)subst) \<Rightarrow> bool"
where "ichain (P,QR) s t \<sigma> = (
 (\<forall>i. ((s i,t i) \<in> P)) \<and> 
 (\<forall>i. ((t i)\<cdot>(\<sigma> i),(s(Suc i))\<cdot>(\<sigma> (Suc i))) \<in> (qrstep QR)^*) \<and>
 (\<forall>i. NF_lhs (fst QR) (s i \<cdot> \<sigma> i))
)"

text {* A \emph{minimal infinite chain} is an infinite chain where additionally all @{text \<sigma>}-instances
of terms in the sequence~@{text t} are terminating w.r.t.~the system~@{text R}. *}
fun min_ichain ::
 "('f,'v)dpp \<Rightarrow> (nat \<Rightarrow> ('f,'v)term) \<Rightarrow> (nat \<Rightarrow> ('f,'v)term) \<Rightarrow> (nat \<Rightarrow> ('f,'v)subst) \<Rightarrow> bool"
where "min_ichain (P,QR) s t \<sigma> = (ichain (P,QR) s t \<sigma> \<and> minimal_cond QR s t \<sigma>)"

definition ID_QR :: "('f,'v)qtrs \<Rightarrow> ('f shp,'v)qtrs"
where
 "ID_QR QR = (plain ` fst QR, ID (snd QR))"


fun initial_dpp :: "('f,'v)qtrs \<Rightarrow> ('f shp,'v)dpp"
where "initial_dpp QR = (DP (useful_rules QR), ID_QR QR)"

definition Tinf :: "('f,'v)trs \<Rightarrow> ('f,'v)term \<Rightarrow> bool"
where "Tinf R t \<equiv> (\<forall>s. t \<rhd> s \<longrightarrow> SN_elt R s) \<and> (\<not> SN_elt R t)"

lemma not_SN_imp_subt_Tinf:
 fixes R :: "('f,'v)trs"
 assumes "\<not> SN_elt R s" shows "\<exists>t. s \<unrhd> t \<and> Tinf R t"
proof -
  let ?S = "{t | t. s \<unrhd> t \<and> \<not> SN_elt R t}"
  from assms have s: "s \<in> ?S" by auto
  from mp[OF spec[OF spec[OF SN_imp_minimal[OF SN_supt]]] s]
  obtain t where st: "s \<unrhd> t" and nSN: "\<not> SN_elt R t" and min: "\<forall> u. (t,u) \<in> supt \<longrightarrow> u \<notin> ?S" by auto
  have "Tinf R t" unfolding Tinf_def 
  proof (simp add: nSN, intro allI impI)
    fix u
    assume u: "t \<unrhd> u \<and> t \<noteq> u" 
    hence tu: "(t,u) \<in> supt" unfolding supt_def by auto
    from u st have "s \<unrhd> u" using supteqp_trans by auto
    with min tu show "SN_elt R u" by auto
  qed
  with st show ?thesis by auto
qed


lemma Tinf_imp_SN_elt_nr_qrstep: assumes Tinf: "Tinf (qrstep QR) t"
  shows "SN_elt (nr_qrstep QR) t"
proof (cases QR)
  case (Pair Q R)
  let ?QR = "(Q,R)"
  have "SN_elt (nr_qrstep ?QR) t"
  proof (cases t)
    case (Var x)
    {
      fix s
      assume "(t,s) \<in> nr_qrstep ?QR"
      then obtain l C \<sigma> where "C \<noteq> \<box>" and "t = C\<langle>l \<cdot> \<sigma>\<rangle>" unfolding nr_qrstep_def rstep_r_c_s_def by auto
      with Var have False by (cases C, auto)
    }
    thus ?thesis unfolding SN_elt_def by auto
  next
    case (Fun f ss)
    with Tinf have SN: "\<And> s. s \<in> set ss \<Longrightarrow> SN_elt (qrstep QR) s" unfolding Tinf_def using suptp.supt_arg by auto
    show ?thesis 
    proof (rule ccontr)
      assume "\<not> ?thesis"
      then obtain S where start: "S 0 = Fun f ss" and steps: "\<And> i. (S i, S (Suc i)) \<in> nr_qrstep ?QR" unfolding SN_elt_def using Fun by force
      let ?P = "\<lambda> i l r g bef C aft \<sigma>. (l,r) \<in> R \<and> S i = Fun g (bef @ C\<langle>l\<cdot>\<sigma>\<rangle> # aft) \<and> S (Suc i) = Fun g (bef @ C\<langle>r\<cdot>\<sigma>\<rangle> # aft) \<and> (\<forall> u. l\<cdot>\<sigma> \<rhd> u \<longrightarrow> NF_lhs Q u)"
      have p: "\<forall> i. \<exists> l r g bef C aft \<sigma>. ?P i l r g bef C aft \<sigma>"
      proof
        fix i
        obtain l r C \<sigma> where "(l,r) \<in> R \<and> C \<noteq> \<box> \<and> S i = C\<langle>l\<cdot>\<sigma>\<rangle> \<and> S (Suc i) = C\<langle>r\<cdot>\<sigma>\<rangle> \<and> (\<forall> u. l\<cdot>\<sigma> \<rhd> u \<longrightarrow> NF_lhs Q u)"
          using steps[of i] unfolding nr_qrstep_def rstep_r_c_s_def by auto
        thus "\<exists> l r g bef C aft \<sigma>. ?P i l r g bef C aft \<sigma>" by (cases C, simp, simp, blast)
      qed
      with choice[OF p] obtain l where p: "\<forall> i. \<exists> r g bef C aft \<sigma>. ?P i (l i) r g bef C aft \<sigma>" by auto
      with choice[OF p] obtain r where p: "\<forall> i. \<exists> g bef C aft \<sigma>. ?P i (l i) (r i) g bef C aft \<sigma>" by auto
      with choice[OF p] obtain g where p: "\<forall> i. \<exists> bef C aft \<sigma>. ?P i (l i) (r i) (g i) bef C aft \<sigma>" by auto
      with choice[OF p] obtain bef where p: "\<forall> i. \<exists> C aft \<sigma>. ?P i (l i) (r i) (g i) (bef i) C aft \<sigma>" by auto
      with choice[OF p] obtain C where p: "\<forall> i. \<exists> aft \<sigma>. ?P i (l i) (r i) (g i) (bef i) (C i) aft \<sigma>" by auto
      with choice[OF p] obtain aft where p: "\<forall> i. \<exists> \<sigma>. ?P i (l i) (r i) (g i) (bef i) (C i) (aft i) \<sigma>" by auto
      with choice[OF p] obtain \<sigma> where p: "\<And> i. ?P i (l i) (r i) (g i) (bef i) (C i) (aft i) (\<sigma> i)" by auto
      let ?gtl = "\<lambda> i. Fun (g i) (bef i @ (C i)\<langle>l i\<cdot>\<sigma> i\<rangle> # aft i)"
      let ?gtr = "\<lambda> i. Fun (g i) (bef i @ (C i)\<langle>r i\<cdot>\<sigma> i\<rangle> # aft i)"
      let ?tl = "\<lambda> i. Fun f (bef i @ (C i)\<langle>l i\<cdot>\<sigma> i\<rangle> # aft i)"
      let ?tr = "\<lambda> i. Fun f (bef i @ (C i)\<langle>r i\<cdot>\<sigma> i\<rangle> # aft i)"
      {
        fix i
        have "?P i (l i) (r i) f (bef i) (C i) (aft i) (\<sigma> i) \<and> Suc (length (bef i) + length (aft i)) = length ss"
        proof (induct i)
          case 0
          from p[of 0] have "?gtl 0 = Fun f ss" using start[symmetric] by auto
          with p[of 0] show ?case by auto
        next
          case (Suc i)
          from p[of i] have "?gtr i = S (Suc i)" by auto
          also have "\<dots> = ?gtl (Suc i)" using p[of "Suc i"] by auto
          finally have "length (args (?gtr i)) = length (args (?gtl (Suc i)))" by auto
          hence "length (bef (Suc i)) + length (aft (Suc i)) = length (bef i) + length (aft i)" by simp          
          with p[of "Suc i"] Suc show ?case by auto
        qed
      } note p = this
      { 
        fix j        
        assume j: "j < length ss"
        {
          fix i
          have "?tl (Suc i) = ?tr i" using p[of i] p[of "Suc i"] by auto
          hence id: "args (?tl (Suc i)) ! j = args (?tr i) ! j" by auto
          have "(args (?tl i) ! j, args (?tl (Suc i)) ! j) \<in> (qrstep ?QR)^="
          proof (cases "j < length (bef i)")
            case True
            hence "args (?tl i) ! j = (bef i) ! j" by (simp add: nth_append)
            moreover from True have "args (?tl (Suc i)) ! j = (bef i) ! j" by (simp only: id, simp add: nth_append)
            ultimately show ?thesis by auto
          next
            case False
            hence jb: "j \<ge> length (bef i)" by auto
            show ?thesis 
            proof (cases "j = length (bef i)")
              case True
              hence "args (?tl i) ! j = (C i)\<langle>l i\<cdot> \<sigma> i\<rangle>" by (simp add: nth_append)              
              moreover have "args (?tl (Suc i)) ! j = (C i)\<langle>r i\<cdot> \<sigma> i\<rangle>" by (simp only: id, simp add: nth_append True)
              ultimately show ?thesis using p[of i] unfolding qrstep_def rstep_r_c_s_def  by auto
            next
              case False
              obtain k where "k = j - (Suc (length (bef i)))" by auto
              with False j jb p[of i] have j: "j = Suc (length (bef i)) + k" and k: "k < length (aft i)" by auto
              hence l: "args (?tl i) ! j = (aft i) ! k" by (simp add: nth_append)              
              moreover have "args (?tl (Suc i)) ! j = (aft i) ! k" by (simp only: id, simp add: nth_append j k)
              ultimately show ?thesis by auto
            qed
          qed
        } note arg_steps = this
        let ?ti = "\<lambda> i. args (?tl i) ! j" 
        from arg_steps  have arg_steps: "\<forall> i. (?ti i, ?ti (Suc i)) \<in> Id \<union> (qrstep ?QR)" by auto
        have compat: "Id O qrstep ?QR \<subseteq> qrstep ?QR" by auto
        from start p[of 0] have "?ti 0 = ss ! j" by auto
        with j have arg: "?ti 0 \<in> set ss" by auto
        hence SN_start: "SN_elt (qrstep ?QR) (?ti 0)" using SN[OF arg] Pair by auto
        have "\<exists> m. \<forall> i \<ge> m. (?ti i, ?ti (Suc i)) \<in> Id - qrstep ?QR" 
          by (rule non_strict_ending[OF arg_steps compat], rule SN_start)
      }
      hence ending: "\<forall> j. \<exists> m. j < length ss \<longrightarrow> (\<forall> i \<ge> m. (args (?tl i) ! j, args (?tl (Suc i)) ! j) \<notin> qrstep ?QR)" by blast
      from choice[OF ending] obtain m where ending: "\<forall> j < length ss. (\<forall> i \<ge> m j. (args (?tl i) ! j, args (?tl (Suc i)) ! j) \<notin> qrstep ?QR)" by blast
      obtain ma where ma: "ma = max_f m (length ss)" by auto
      hence ma: "\<forall> i < length ss. m i \<le> ma" using max_f by auto
      with ending have ending: "\<forall> j < length ss. (\<forall> i \<ge> ma. (args (?tl i) ! j, args (?tl (Suc i)) ! j) \<notin> qrstep ?QR)" by auto
      let ?l = "length (bef ma)"
      from p[of ma] have len: "?l < length ss" by auto
      from p[of ma] have l: "args (?tl ma) ! ?l = (C ma)\<langle>l ma \<cdot> \<sigma> ma\<rangle>" by (simp add: nth_append)
      from p[of ma] p[of "Suc ma"] have "?tl (Suc ma) = ?tr ma" by auto
      hence "args (?tl (Suc ma)) ! ?l = args (?tr ma) ! ?l" by auto
      also have "\<dots> = (C ma)\<langle>r ma \<cdot> \<sigma> ma\<rangle>" using p[of ma] by (simp add: nth_append)
      finally have "(args (?tl ma) ! ?l, args (?tl (Suc ma)) ! ?l) \<in> qrstep ?QR" using p[of ma] by (simp only: l, unfold qrstep_def rstep_r_c_s_def, auto)
      with len ending show False by blast
    qed
  qed
  with Pair show ?thesis by simp
qed


lemma NF_lhs_iff_NF_lhs_plain: "NF_lhs (plain ` Q) (plain s) = NF_lhs Q s"
proof -
  obtain R where R: "R = {(q,q) | q. q \<in> Q}" by auto
  hence QR: "Q = fst ` R \<and> plain ` Q = fst ` ID R" unfolding ID_def by force
  have "NF_lhs (fst ` R) s = NF_lhs (fst ` ID R) (plain s)" 
  proof (simp only: NF_lhs_NF_rstep_conv, simp only: NF_iff_no_step)
    have "(\<forall> u. (s,u) \<notin> rstep R) = (\<forall> u. (plain s, plain u) \<notin> rstep (ID R))"
      by (simp only: rstep_iff_plain_rstep[symmetric])
    also have "\<dots> = (\<forall> u. (plain s, u) \<notin> rstep (ID R))"
    proof -
      {
        fix u
        assume "(plain s, u) \<in> rstep (ID R)"
        then obtain l C \<sigma> where "plain s = C\<langle>plain l\<cdot> \<sigma>\<rangle> \<and> u = C\<langle>plain l\<cdot>\<sigma>\<rangle>" 
          unfolding rstep_iff_rstep_r_c_s rstep_r_c_s_def R ID_def by auto
        hence "u = plain s" by auto
      }
      thus ?thesis by auto 
    qed
    finally show "(\<forall> u. (s,u) \<notin> rstep R) = (\<forall> u. (plain s, u) \<notin> rstep (ID R))" .
  qed
  with QR show ?thesis by auto
qed

lemma plain_imp_supt_plain: assumes "plain s \<rhd> t" shows "\<exists> u. t = plain u \<and> s \<rhd> u"  
proof -
  from assms have sub: "plain s \<unrhd> t" and nid: "plain s \<noteq> t" by auto
  from subteq_of_plain_imp_plain[OF sub] obtain u where t: "t = plain u" by auto
  with assms have "plain s \<rhd> plain u" by auto
  hence "s \<rhd> u" by (rule suptp_plain')
  with t show ?thesis by auto
qed

lemma qrstep_iff_plain_qrstep: "((plain s,plain t) \<in> qrstep (ID_QR QR)) = ((s,t) \<in> qrstep QR)"
proof (cases QR)
  case (Pair Q R)
  let ?QR = "(Q,R)"
  let ?qr = "qrstep QR"
  let ?IQR = "ID_QR ?QR"
  let ?Iqr = "qrstep ?IQR"
  let ?ps = "\<lambda> s. Subst (plain \<circ> (case s of Subst si \<Rightarrow> si))"
  have "((s,t) \<in> ?qr) = ((plain s, plain t) \<in> ?Iqr)"
  proof
    assume "(s,t) \<in> ?qr"
    then obtain l r C \<sigma> where "(l,r) \<in> R" and "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and "t = C\<langle>r \<cdot> \<sigma>\<rangle>" and NF: "\<forall> u. l \<cdot> \<sigma> \<rhd> u \<longrightarrow> NF_lhs Q u" unfolding Pair qrstep_def rstep_r_c_s_def by auto
    hence rule: "(plain l, plain r) \<in> ID R" and step: "(plain s, plain t) \<in> rstep_r_c_s (plain l, plain r) (plain_ctxt C) (?ps \<sigma>)"
      using plain_subst_distr plain_ctxt_distr unfolding ID_def rstep_r_c_s_def by auto
    have "\<forall> u. plain (l \<cdot> \<sigma>) \<rhd> u \<longrightarrow> NF_lhs (plain ` Q) u"
    proof (intro impI allI)
      fix u
      assume pl: "plain (l \<cdot> \<sigma>) \<rhd> u"
      from plain_imp_supt_plain[OF pl] obtain v where u: "u = plain v" and v: "l \<cdot> \<sigma> \<rhd> v" by auto
      from v NF have "NF_lhs Q v" by auto
      hence "NF_lhs (plain ` Q) (plain v)" by (simp only: NF_lhs_iff_NF_lhs_plain)
      with u show "NF_lhs (plain ` Q) u" by simp
    qed
    with rule step show "(plain s, plain t) \<in> ?Iqr" unfolding qrstep_def ID_QR_def by auto
  next
    let ?us = "\<lambda> s. Subst (unsharp \<circ> (case s of Subst si \<Rightarrow> si))"
    assume "(plain s, plain t) \<in> ?Iqr"
    then obtain l r C \<sigma> where "(l,r) \<in> R" and s: "plain s = C\<langle>plain l \<cdot> \<sigma>\<rangle>" and "plain t = C\<langle>plain r \<cdot> \<sigma>\<rangle>" and NF: "\<forall> u. plain l \<cdot> \<sigma> \<rhd> u \<longrightarrow> NF_lhs (plain ` Q) u"
      unfolding qrstep_def ID_def rstep_r_c_s_def ID_QR_def by auto
    hence "unsharp(plain s) = unsharp(C\<langle>plain l \<cdot> \<sigma>\<rangle>)" and "unsharp(plain t) = unsharp(C\<langle>plain r \<cdot> \<sigma>\<rangle>)" by auto
    hence step: "(s,t) \<in> rstep_r_c_s (l,r) (unsharp_ctxt C) (?us \<sigma>)" unfolding rstep_r_c_s_def 
      using unsharp_subst_distr unsharp_ctxt_distr by auto
    from s have "plain s \<unrhd> plain l \<cdot> \<sigma>" using ctxt_imp_supteqp by auto
    with subteq_of_plain_imp_plain obtain ll where ll: "plain l \<cdot> \<sigma> = plain ll" by auto
    with NF have NF: "\<forall> u. plain ll \<rhd> u \<longrightarrow> NF_lhs (plain ` Q) u" by auto
    have "l \<cdot> ?us \<sigma> = unsharp (plain l) \<cdot> ?us \<sigma>" by auto
    also have "\<dots> = unsharp (plain l \<cdot> \<sigma>)" using unsharp_subst_distr by auto
    finally have ll2: "l \<cdot> ?us \<sigma> = ll" by (simp add: ll)
    have "\<forall> u. l \<cdot> ?us \<sigma> \<rhd> u \<longrightarrow> NF_lhs Q u"
    proof (intro allI impI, simp only: ll2)
      fix u
      assume "ll \<rhd> u"
      hence "plain ll \<rhd> plain u" by (rule suptp_plain)
      with NF have "NF_lhs (plain ` Q) (plain u)" by auto 
      with NF_lhs_iff_NF_lhs_plain show "NF_lhs Q u" by auto
    qed
    with step `(l,r) \<in> R` show "(s,t) \<in> ?qr" unfolding qrstep_def Pair by auto
  qed
  with Pair show ?thesis by auto
qed

lemma useful_plain_rule_imp_useful_rule: assumes "useful_rule (plain ` Q) (plain l, plain r)" shows "useful_rule Q (l,r)"
proof (rule ccontr)
  assume "\<not> ?thesis"
  then obtain s q \<sigma> where supt: "l \<rhd> s" and q: "q \<in> Q" and s: "s = q \<cdot> \<sigma>" unfolding useful_rule_def by auto
  hence "fst (plain l,plain r) \<rhd> plain s" and "plain q \<in> plain ` Q" and "plain s = plain q \<cdot> (Subst (plain \<circ> (case \<sigma> of Subst si \<Rightarrow> si)))"
    using suptp_plain[OF supt] plain_subst_distr by auto
  with assms show False unfolding useful_rule_def by  blast
qed
    
  
lemma ID_QR_preserves_plain_terms: assumes SN: "SN_elt (qrstep QR) s" and step: "(plain s, t) \<in> qrstep (ID_QR QR)"
  shows "plain(unsharp t) = t"
proof (cases QR)
  case (Pair Q R)
  let ?QR = "(Q,R)"
  let ?IQR = "ID_QR ?QR"
  let ?pQ = "plain ` Q"
  from step obtain l r C \<sigma> where ps: "plain s = C\<langle>l \<cdot> \<sigma>\<rangle>" and lr: "(l,r) \<in> ID R" and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>" and NF: "\<forall> u. l \<cdot> \<sigma> \<rhd> u \<longrightarrow> NF_lhs ?pQ u"
    unfolding ID_QR_def qrstep_def rstep_r_c_s_def by (simp add: Pair, blast)
  from only_useful_rules[OF NF] have uselr: "useful_rule (plain ` Q) (l,r)" by auto
  from lr obtain ll rr where "l = plain ll" and "r = plain rr" and "(ll,rr) \<in> R" unfolding ID_def by auto
  hence l: "plain(unsharp l) = l" and r: "plain (unsharp r) = r" and ulur: "(unsharp l, unsharp r) \<in> R" by auto
  let ?ul = "unsharp l" 
  let ?ur = "unsharp r"
  have useulur: "useful_rule Q (?ul, ?ur)" by (rule useful_plain_rule_imp_useful_rule, simp add: l r uselr)
  let ?Cls = "C \<langle>l \<cdot> \<sigma>\<rangle>"
  let ?C = "plain_ctxt (unsharp_ctxt C)"
  let ?us = "Subst (unsharp \<circ> (case \<sigma> of Subst si \<Rightarrow> si))"
  from ctxt_imp_supteqp ps have "plain s \<unrhd> l \<cdot> \<sigma>" by auto
  with subteq_of_plain_imp_plain obtain tmp where "l \<cdot> \<sigma> = plain tmp" by auto
  hence ls: "plain (unsharp (l \<cdot> \<sigma>)) = l \<cdot> \<sigma>" by auto
  have "s = unsharp (plain s)" by simp
  also have "\<dots> = unsharp (?Cls)" by (simp add: ps)
  also have "\<dots> = (unsharp_ctxt C)\<langle>?ul \<cdot> ?us\<rangle>" by simp
  finally have s: "s = (unsharp_ctxt C)\<langle>?ul \<cdot> ?us\<rangle>" .
  have NF2: "\<forall> u. ?ul \<cdot> ?us \<rhd> u \<longrightarrow> NF_lhs Q u" 
  proof (intro allI impI)
    fix u
    assume "?ul \<cdot> ?us \<rhd> u"
    hence "unsharp (l \<cdot> \<sigma>) \<rhd> u" by simp
    hence "plain (unsharp (l \<cdot> \<sigma>)) \<rhd> plain u" by (rule suptp_plain)
    hence "l \<cdot> \<sigma> \<rhd> plain u" by (simp only: ls)
    with NF have "NF_lhs ?pQ (plain u)" by auto
    thus "NF_lhs Q u" unfolding NF_lhs_iff_NF_lhs_plain  .
  qed
  with SN_elt_imp_wwf_rule[OF SN[simplified Pair] s ulur] have "wwf_rule Q (?ul,?ur)" by simp
  with useulur have "vars_term ?ul \<supseteq> vars_term ?ur" unfolding wwf_rule_def by auto
  hence vars: "vars_term l \<supseteq> vars_term r" using vars_plain_eq_vars l r by force 
  from plain_ctxt_term[OF ps] have C: "?C = C" by simp
  let ?sl = "\<sigma>|_vars_term l"
  let ?pus = "Subst (plain \<circ> unsharp \<circ> (case \<sigma> of Subst si \<Rightarrow> si))"
  let ?pusl = "Subst (plain \<circ> unsharp \<circ> (case ?sl of Subst s \<Rightarrow> s))"
  from plain_subst_term[OF ls, simplified coincidence_lemma[symmetric]] have "l \<cdot> \<sigma> = l \<cdot> ?pusl" ..
  hence "r \<cdot> \<sigma> = r \<cdot> ?pusl" by (rule vars_subst_ex[OF vars])
  also have "\<dots> = r \<cdot> ?pus"
  proof -
    from vars show "r \<cdot> ?pusl = r \<cdot> ?pus"
    proof (induct r)
      case (Fun f ss)
      thus ?case by auto
    next
      case (Var x)
      thus ?case by (cases \<sigma>, simp)
    qed
  qed      
  finally have rs: "r \<cdot> ?pus = r \<cdot> \<sigma>" by simp
  have "plain (unsharp t) = plain (unsharp (C\<langle>r \<cdot> \<sigma>\<rangle>))" by (simp add: t)
  also have "\<dots>  = C\<langle>r \<cdot> \<sigma>\<rangle>" by (simp add: C r o_assoc rs)
  finally show ?thesis by (simp add: t)
qed 

lemma qrstepID_imp_plain_qrstep: assumes SN: "SN_elt (qrstep QR) s" and step: "(plain s,t) \<in> qrstep (ID_QR QR)"
  shows "(plain(s), plain(unsharp t)) \<in> qrstep (ID_QR QR)"
proof -
  from step show ?thesis unfolding ID_QR_preserves_plain_terms[OF SN step] . 
qed
    

lemma SN_elt_qrstep_iff_SN_elt_qrstepID:
 fixes QR :: "('f,'v)qtrs" and s :: "('f,'v)term"
 shows "(SN_elt (qrstep (ID_QR QR)) (plain s)) = (SN_elt (qrstep QR) s)" (is "?R = ?L")
proof (cases QR)
  case (Pair Q R)
  let ?QR = "(Q,R)"
  let ?qr = "qrstep ?QR"
  let ?IQR = "ID_QR ?QR"
  let ?Iqr = "qrstep ?IQR"
  have "SN_elt ?qr s = SN_elt ?Iqr (plain s)" 
  proof
    assume SN: "SN_elt ?qr s"
    show "SN_elt ?Iqr (plain s)"
    proof (rule ccontr)
      assume "\<not>?thesis"
      then obtain S where start: "S 0 = plain s" and seq: "\<forall>i. (S i,S(Suc i)) \<in> ?Iqr"
        unfolding SN_elt_def by auto
      let ?S = "\<lambda>i. plain(unsharp(S i))"
      let ?T = "\<lambda>i. unsharp(S i)"
      have "\<forall> i. ?S (Suc i) = S (Suc i) \<and> (?T i, ?T(Suc i)) \<in> ?qr \<and> SN_elt ?qr (?T (Suc i))"
      proof
        fix i
        show "?S (Suc i) = S (Suc i) \<and> (?T i, ?T(Suc i)) \<in> ?qr \<and> SN_elt ?qr (?T (Suc i))"
        proof (induct i)
          case 0
          from start have id: "?S 0 = S 0" by auto
          from start SN have SN: "SN_elt ?qr (?T 0)" by auto
          from ID_QR_preserves_plain_terms[OF SN, unfolded id, OF spec[OF seq, of 0]]
          have id2: "?S (Suc 0) = S (Suc 0)" . 
          from spec[OF seq, of 0] have "(?S 0, ?S (Suc 0)) \<in> ?Iqr" unfolding id id2 .
          hence step: "(?T 0, ?T (Suc 0)) \<in> ?qr" unfolding qrstep_iff_plain_qrstep .
          hence "(?T 0, ?T (Suc 0)) \<in> ?qr^*" by simp
          hence "SN_elt ?qr (?T (Suc 0))" by (rule steps_preserve_SN_elt[OF _ SN])
          with id2 step show ?case by auto
        next
          case (Suc i)
          let ?i = "Suc i"
          from Suc have id: "?S ?i = S ?i" 
           and SN: "SN_elt ?qr (?T ?i)" by auto
          from ID_QR_preserves_plain_terms[OF SN, unfolded id, OF spec[OF seq, of ?i]]
          have id2: "?S (Suc ?i) = S (Suc ?i)" . 
          from spec[OF seq, of ?i] have "(?S ?i, ?S (Suc ?i)) \<in> ?Iqr" unfolding id id2 .
          hence step: "(?T ?i, ?T (Suc ?i)) \<in> ?qr" unfolding qrstep_iff_plain_qrstep .
          hence "(?T ?i, ?T (Suc ?i)) \<in> ?qr^*" by simp
          hence "SN_elt ?qr (?T (Suc ?i))" by (rule steps_preserve_SN_elt[OF _ SN])
          with id2 step show ?case by auto
        qed
      qed
      hence seq': "\<forall>i. (?T i,?T(Suc i)) \<in> ?qr" by blast
      hence "?T 0 = s" unfolding `S 0 = plain s` by simp
      with seq' have "\<not>(SN_elt ?qr s)" unfolding SN_elt_def by auto
      with `SN_elt ?qr s` show False by simp
    qed
  next
    assume "SN_elt ?Iqr (plain s)"
    show "SN_elt ?qr s"
    proof (rule ccontr)
      assume "\<not>?thesis"
      then obtain S where start: "S 0 = s" and seq: "\<forall>i. (S i,S(Suc i)) \<in> ?qr"
        unfolding SN_elt_def by auto
      let ?S = "\<lambda>i. plain(S i)"
      from seq have seq': "\<forall>i. (?S i,?S(Suc i)) \<in> ?Iqr" unfolding qrstep_iff_plain_qrstep .
      hence "?S 0 = plain s" unfolding `S 0 = s` by simp
      with seq' have "\<not>(SN_elt ?Iqr (plain s))" unfolding SN_elt_def by auto
      with `SN_elt ?Iqr (plain s)` show False by simp
    qed
  qed
  with Pair show ?thesis by auto
qed
   

lemma wwf_rule_imp_wwf_rule_ID: assumes "wwf_rule Q (l,r)" shows "wwf_rule (plain ` Q) (plain l, plain r)"
unfolding wwf_rule_def
proof (simp, intro impI)
  let ?Q = "plain ` Q"
  let ?l = "plain l"
  let ?r = "plain r"
  assume "useful_rule ?Q (?l,?r)"
  hence "useful_rule Q (l,r)" by (rule useful_plain_rule_imp_useful_rule)
  with assms obtain f ts where l: "l = Fun f ts" and v: "vars_term r \<subseteq> vars_term l" unfolding wwf_rule_def by blast
  from l have  pl: "?l = Fun (Plain f) (map plain ts)" by auto
  from v have pv: "vars_term ?r \<subseteq> vars_term ?l" (is ?sub) by (simp add: vars_plain_eq_vars)
  from pl pv show "(\<exists> f ts. ?l = Fun f ts) \<and> ?sub" by auto
qed
  
lemma wwf_qtrs_imp_wwf_qtrs_ID: assumes "wwf_qtrs QR" shows "wwf_qtrs (ID_QR QR)"
  using assms
  by (cases QR, simp add: ID_QR_def ID_def, auto simp: wwf_qtrs_def wwf_rule_imp_wwf_rule_ID)


lemma Tinf_imp_SN_sharp:
 fixes QR :: "('f,'v)qtrs"
 assumes "wwf_qtrs QR" and "Tinf (qrstep QR) s"
 shows "SN_elt (qrstep(ID_QR QR)) (sharp s)"
proof (cases QR)
  case (Pair Q R)
  let ?QR = "(Q,R)"
  let ?qr = "qrstep ?QR"
  let ?IQR = "ID_QR ?QR"
  let ?Iqr = "qrstep ?IQR"
  from Pair assms have wf: "wwf_qtrs ?QR" and Tinf: "Tinf ?qr s"  by auto
  have "SN_elt ?Iqr (sharp s)" 
  proof (rule ccontr)
    assume nontermin: "\<not>(SN_elt ?Iqr (sharp s))"
    then obtain S where "S 0 = sharp s" and seq:"\<forall>i. (S i,S(Suc i)) \<in> ?Iqr" unfolding SN_elt_def by auto
    hence union: "\<forall>i. (S i,S(Suc i)) \<in> r_qrstep ?IQR \<union> nr_qrstep ?IQR" unfolding qrstep_iff_r_qrstep_or_nr_qrstep by simp
    from seq have step: "(sharp s,S(Suc 0)) \<in> ?Iqr" unfolding `S 0 = sharp s`[symmetric] by simp
    then obtain l r C \<sigma> where step': "(sharp s,S(Suc 0)) \<in> rstep_r_c_s (l,r) C \<sigma>" and "(l,r) \<in> ID R" unfolding qrstep_def ID_QR_def by auto
    hence "C\<langle>l\<cdot>\<sigma>\<rangle> = sharp s" and "C\<langle>r\<cdot>\<sigma>\<rangle> = S(Suc 0)" unfolding rstep_r_c_s_def by auto
    have wfi: "wwf_qtrs (plain ` Q, ID R)" using wwf_qtrs_imp_wwf_qtrs_ID[OF wf] unfolding ID_QR_def by auto
    from `(l,r) \<in> ID R` obtain g ts where "l = Fun g ts" using wwf_qtrs_imp_left_fun[OF wfi] by best
    obtain f ss where "s = Fun f ss" using `C\<langle>l\<cdot>\<sigma>\<rangle> = sharp s` unfolding `l = Fun g ts` by (induct s,cases C) auto
    have "sharp s = Fun (Sharp f) (map plain ss)" (is "_ = Fun _ ?ss") unfolding `s = Fun f ss` by simp
    hence "S 0 = Fun (Sharp f) ?ss" unfolding `S 0 = sharp s` by simp
    have "\<forall>u. sharp s \<rhd> u \<longrightarrow> SN_elt ?Iqr u"
    proof (intro allI impI)
      fix u assume "sharp s \<rhd> u"
      then obtain v where "u = plain v" using subt_of_sharp_imp_plain by auto
      from `sharp s \<rhd> u` have "s \<rhd> v" unfolding `u = plain v` suptp_sharp_conv by simp
      with `Tinf ?qr s` have "SN_elt ?qr v" unfolding Tinf_def by auto
      thus "SN_elt ?Iqr u" unfolding `u = plain v`  using SN_elt_qrstep_iff_SN_elt_qrstepID by auto
    qed
    with nontermin have "Tinf ?Iqr (sharp s)" unfolding Tinf_def by auto
    hence "SN_elt (nr_qrstep ?IQR) (sharp s)" by (rule Tinf_imp_SN_elt_nr_qrstep)
    hence first: "SN_elt (nr_qrstep ?IQR) (S 0)" unfolding `S 0 = sharp s` by simp
    obtain i where "(S i,S(Suc i)) \<in> r_qrstep ?IQR"
      and prefix: "\<forall>j<i. (S j,S(Suc j)) \<in> nr_qrstep ?IQR \<and> (S j,S(Suc j)) \<notin> r_qrstep ?IQR"
      using union_iseq_SN_elt_imp_first_step[where t = S,OF union first] by auto
    have less: "\<forall>j<i.\<exists>us. S j = Fun (Sharp f) us"
    proof (intro allI impI)
      fix j assume "j < i" thus "\<exists>us. S j = Fun (Sharp f) us"
      proof (induct j)
        case 0 thus ?case using `S 0 = Fun (Sharp f) ?ss` by best
      next
        case (Suc j)
        then obtain us' where "S j = Fun (Sharp f) us'" by auto
        have "(S j,S(Suc j)) \<in> nr_qrstep ?IQR" using prefix and Suc by auto
        then obtain us'' where "S(Suc j) = Fun (Sharp f) us''" using nr_qrstep_preserves_root unfolding `S j = Fun (Sharp f) us'` by best
        thus ?case by auto
      qed
    qed
    have "\<exists>us. S i = Fun (Sharp f) us"
    proof (cases i)
      case 0 thus ?thesis using `S 0 = Fun (Sharp f) ?ss` by auto
    next
      case (Suc n)
      from less have "\<exists>us. S n = Fun (Sharp f) us" unfolding Suc by auto
      then obtain vs where "S n = Fun (Sharp f) vs" by auto
      have "(S n,S(Suc n)) \<in> nr_qrstep ?IQR" using prefix unfolding Suc by auto
      then obtain vs' where "S i = Fun (Sharp f) vs'" using nr_qrstep_preserves_root unfolding Suc `S n = Fun (Sharp f) vs` by fast
      thus ?thesis by auto
    qed
    then obtain us where "S i = Fun (Sharp f) us" by auto
    from `(S i,S(Suc i)) \<in> r_qrstep ?IQR` have "(Fun (Sharp f) us,S(Suc i)) \<in> r_qrstep ?IQR" unfolding `S i = Fun (Sharp f) us` by auto
    then obtain l r \<sigma> where "(l,r) \<in> ID R" and "l\<cdot>\<sigma> = Fun (Sharp f) us" unfolding r_qrstep_def rstep_r_c_s_def ID_QR_def by auto
    then obtain l' r' where "(l',r') \<in> R" and "l = plain l'" unfolding ID_def by auto
    with `wwf_qtrs ?QR` obtain g vs where "l' = Fun g vs" using wwf_qtrs_imp_left_fun by best
    from `l\<cdot>\<sigma> = Fun (Sharp f) us` show False unfolding `l = plain l'` `l' = Fun g vs` by simp
  qed
  with Pair show ?thesis by simp
qed




lemma Tinf_lemma_1:
  assumes "wwf_qtrs (Q,R)" and "Tinf (qrstep (Q,R)) (Fun f ts)" (is "Tinf (qrstep ?QR) ?t")
  shows "\<exists>l r u \<sigma>.
  ((l,r) \<in> useful_rules ?QR) \<and> 
  (r \<unrhd> u) \<and> 
  (\<forall>x. u \<noteq> (Var x)) \<and> 
  ((?t,l\<cdot>\<sigma>) \<in> (nr_qrstep ?QR)^*) \<and> 
  ((l\<cdot>\<sigma>,r\<cdot>\<sigma>) \<in> r_qrstep ?QR) \<and> 
  ((r\<cdot>\<sigma>) \<unrhd> (u\<cdot>\<sigma>)) \<and>
  (Tinf (qrstep (Q,R)) (l\<cdot>\<sigma>)) \<and>
  (Tinf (qrstep (Q,R)) (u\<cdot>\<sigma>)) \<and>
  \<not>(l \<rhd> u)"
proof -
  let ?U = "useful_rules ?QR"
  let ?qr = "qrstep ?QR"
  let ?rqr = "r_qrstep ?QR"
  let ?nrqr = "nr_qrstep ?QR"
  from assms have "!!s. ?t \<rhd> s \<Longrightarrow> SN_elt ?qr s" and "\<not> SN_elt ?qr ?t" unfolding Tinf_def by auto
  from `\<not> SN_elt ?qr ?t` have "\<exists>A. A 0 = ?t \<and> (\<forall>i. (A i,A(Suc i)) \<in> ?qr)"
    unfolding SN_elt_def by auto
  then obtain "A" where "A 0 = ?t" and "\<forall>i. (A i,A(Suc i)) \<in> ?qr" by auto
  hence "!!i. (A i,A(Suc i)) \<in> (?rqr \<union> ?nrqr)" by (auto simp: qrstep_iff_r_qrstep_or_nr_qrstep)
  hence "(A 0,A 1) \<in> ?rqr \<or> (A 0,A 1) \<in> ?nrqr" by force
  hence "\<exists>i. (A 0,A i) \<in> ?nrqr^* \<and> (A i,A(Suc i)) \<in> ?rqr"
  proof
    assume A: "(A 0,A 1) \<in> ?rqr"
    have "(A 0,A 0) \<in> ?nrqr^*" by auto
    with A show ?thesis by auto
  next
    assume B: "(A 0,A 1) \<in> ?nrqr"
    from `Tinf ?qr ?t` have "SN_elt ?nrqr ?t" by (rule Tinf_imp_SN_elt_nr_qrstep)
    from `A 0 = ?t`[symmetric] `SN_elt ?nrqr ?t` have "SN_elt ?nrqr (A 0)" by simp
    from `!!i. (A i,A(Suc i)) \<in> (?rqr \<union> ?nrqr)` `SN_elt ?nrqr (A 0)`
    have "\<exists>i. (A i,A(Suc i)) \<in> ?rqr \<and> (\<forall>j<i. (A j,A(Suc j)) \<in> ?nrqr
      \<and> \<not> (A j,A(Suc j)) \<in> ?rqr)"
      using union_iseq_SN_elt_imp_first_step[where R = ?rqr and S = ?nrqr and t = A] by force
    then obtain i where first_rstep: "(A i,A(Suc i)) \<in> ?rqr"
      and "!!j. j<i \<Longrightarrow> (A j,A(Suc j)) \<in> ?nrqr \<and> \<not> (A j,A(Suc j)) \<in> ?rqr" by auto
    hence "\<forall>j<i. (A j,A(Suc j)) \<in> ?nrqr" by auto
    hence "(A 0,A i) \<in> ?nrqr^*" by (induct i) auto
    with first_rstep show ?thesis by auto
  qed
  then obtain i where "(A 0,A i) \<in> ?nrqr^*" and step: "(A i,A(Suc i)) \<in> ?rqr" by auto
  then obtain l r \<sigma> where "(l,r) \<in> R" and "l\<cdot>\<sigma> = A i" and "r\<cdot>\<sigma> = A (Suc i)" and NF: "\<forall> v. l \<cdot> \<sigma> \<rhd> v \<longrightarrow> NF_lhs Q v" unfolding r_qrstep_def rstep_r_c_s_def by auto
  with step have "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> ?rqr" by auto
  from `(l,r) \<in> R` only_useful_rules[OF NF] have "(l,r) \<in> ?U" unfolding useful_rules_def by auto
  from `A 0 = ?t` and `(A 0,A i) \<in> ?nrqr^*` have "(?t,A i) \<in> ?nrqr^*" by simp
  hence "\<exists>ls. A i = Fun f ls" by (rule nr_qrstep_seq_preserves_root)
  then obtain ls' where "A i = Fun f ls'" by auto
  have A: "!!i. i < num_args ?t \<Longrightarrow> (ts!i,ls'!i) \<in> ?qr^*" (is "!!i. i < num_args ?t \<Longrightarrow> ?P i")
  proof -
    fix j
    assume "j < num_args ?t"
    with `(?t,A i) \<in> ?nrqr^*` have "(args ?t!j,args (Fun f ls')!j) \<in> ?qr^*"
      unfolding `A i = Fun f ls'` by (rule nr_qrstep_seq_imp_arg_qrstep_seq)
    thus "?P j" by auto
  qed
  from `(?t,A i) \<in> ?nrqr^*` and `A i = Fun f ls'` have "(?t,Fun f ls') \<in> ?nrqr^*" by simp
  from `l\<cdot>\<sigma> = A i` and `A i = Fun f ls'` have "l\<cdot>\<sigma> = Fun f ls'" by simp
  with `(?t,Fun f ls') \<in> ?nrqr^*` have "(?t,l\<cdot>\<sigma>) \<in> ?nrqr^*" by simp
  from `(?t,Fun f ls') \<in> ?nrqr^*` have numargs: "num_args ?t = num_args (Fun f ls')"
    by (rule nr_qrstep_seq_equiv_num_args)
  from `wwf_qtrs ?QR` and `(l,r) \<in> R` obtain g ls where "l = Fun g ls"
    using wwf_qtrs_imp_left_fun by best
  hence "l\<cdot>\<sigma> = Fun g (map (\<lambda>t. t\<cdot>\<sigma>) ls)" by simp
  with `l\<cdot>\<sigma> = Fun f ls'` have "g = f" and "ls' = map (\<lambda>t. t\<cdot>\<sigma>) ls" by auto
  hence B: "!!i. i < length ls' \<Longrightarrow> ls'!i = (ls!i)\<cdot>\<sigma>" by auto
  with A have "!!i. i < num_args ?t \<Longrightarrow> (ts!i,(ls!i)\<cdot>\<sigma>) \<in> ?qr^*"
  proof -
    fix i assume "i < num_args ?t" "!!i. i < num_args ?t \<Longrightarrow> (ts!i,ls'!i) \<in> ?qr^*"
      and "!!i. i < length ls' \<Longrightarrow> ls'!i = (ls!i)\<cdot>\<sigma>"
    hence "(ts!i,ls'!i) \<in> ?qr^*" by simp
    from numargs `i < num_args ?t` have "i < length ls'" by simp
    with `!! i. i < length ls' \<Longrightarrow> ls'!i = (ls!i)\<cdot>\<sigma>` and `(ts!i,ls'!i) \<in> ?qr^*`
    show "(ts!i,(ls!i)\<cdot>\<sigma>) \<in> ?qr^*" by simp
  qed 
  from `l \<cdot> \<sigma> = Fun f ls'` numargs have "num_args ?t = num_args (l\<cdot>\<sigma>)"  by simp
  with `l\<cdot>\<sigma> = Fun f ls'` and `!!i. i < length ls' \<Longrightarrow> ls'!i = (ls!i)\<cdot>\<sigma>`
  have "!!i. i < num_args ?t \<Longrightarrow> (ls!i)\<cdot>\<sigma> = ls'!i" by auto
  have "!!i. i < num_args ?t \<Longrightarrow> ?t \<rhd> (ts!i)"
  proof -
    fix i assume "i < num_args ?t"
    hence "ts!i \<in> set ts" by simp
    thus "?t \<rhd> (ts!i)" by (rule suptp.supt_arg)
  qed
  with `!!s. ?t \<rhd> s \<Longrightarrow> SN_elt ?qr s` have "!!i. i < num_args ?t \<Longrightarrow> SN_elt ?qr (ts!i)" by auto
  with `!!i. i < num_args ?t \<Longrightarrow> (ts!i,(ls!i)\<cdot>\<sigma>) \<in> ?qr^*` 
  have "!!i. i < num_args ?t \<Longrightarrow> SN_elt ?qr ((ls!i)\<cdot>\<sigma>)" using steps_preserve_SN_elt[of _ _ ?qr] by blast
  with `!!i. i < num_args ?t \<Longrightarrow> (ls!i)\<cdot>\<sigma> = ls'!i` and `num_args ?t = num_args (l\<cdot>\<sigma>)`
    and `l\<cdot>\<sigma> = Fun f ls'` and `!!i. i < num_args ?t \<Longrightarrow> SN_elt ?qr ((ls!i)\<cdot>\<sigma>)`
  have argsSN: "!!i. i < num_args (Fun f ls') \<Longrightarrow> SN_elt ?qr (ls'!i)" by auto
  have "\<forall>s. Fun f ls' \<rhd> s \<longrightarrow> SN_elt ?qr s" 
  proof (intro allI impI)
    fix s
    assume tmp: "Fun f ls' \<rhd> s"
    from  suptp_imp_nectxt[OF tmp] obtain C where "C \<noteq> \<box>" and "Fun f ls' = C\<langle>s\<rangle>" by auto
    then obtain bef aft C where id: "Fun f ls' = Fun f (bef @ C\<langle>s\<rangle> # aft)" by (cases C, auto)
    hence len: "length bef < num_args (Fun f ls')" by auto
    with argsSN[OF len] id have "SN_elt ?qr (C \<langle>s\<rangle>)" by auto
    with SN_rel_imp_SN_subt[OF ctxt_closed_qrstep _ ctxt_imp_supteqp[of C s]]
    show "SN_elt ?qr s" by auto
  qed
  hence "\<forall>s. l\<cdot>\<sigma> \<rhd> s \<longrightarrow> SN_elt ?qr s" unfolding `l\<cdot>\<sigma> = Fun f ls'` by simp 
  let ?B = "\<lambda>j. A (i + j)"
  have "?B 0 = l\<cdot>\<sigma>" unfolding `l\<cdot>\<sigma> = A i` by simp
  have "\<forall>i. (?B i,?B(Suc i)) \<in> ?qr" using `\<forall>i. (A i,A(Suc i)) \<in> ?qr` by simp
  with `?B 0 = l\<cdot>\<sigma>` have "\<not> SN_elt ?qr (l\<cdot>\<sigma>)" unfolding SN_elt_def by best
  with `\<forall>s. l\<cdot>\<sigma> \<rhd> s \<longrightarrow> SN_elt ?qr s` have tinf_ls: "Tinf ?qr (l\<cdot>\<sigma>)" unfolding Tinf_def by simp
  have "\<And> x. x \<in> vars_term r \<Longrightarrow> l \<rhd> Var x"
  proof -
    fix x
    assume "x \<in> vars_term r" hence "x \<in> vars_term l" using `(l,r) \<in> ?U` and `wwf_qtrs ?QR` unfolding wwf_qtrs_def wwf_rule_def useful_rules_def by auto
    thus "l \<rhd> Var x" using `l = Fun g ls` supteqp_var[OF `x \<in> vars_term l`] by auto
  qed
  hence "\<And> x. x \<in> vars_term r \<Longrightarrow> l \<cdot> \<sigma> \<rhd> Var x \<cdot> \<sigma>" using suptp_stable[of l] by best
  hence xSN: "\<And> x. x \<in> vars_term r \<Longrightarrow> SN_elt ?qr (Var x \<cdot> \<sigma>)" using tinf_ls unfolding Tinf_def by auto
  let ?B = "\<lambda>j. A (Suc i + j)"
  from `r\<cdot>\<sigma> = A (Suc i)` have "?B 0 = r\<cdot>\<sigma>" by simp
  from `\<forall>i. (A i,A(Suc i)) \<in> ?qr` have "\<forall>i. (?B i,?B(Suc i)) \<in> ?qr" by auto
  with `?B 0 = r\<cdot>\<sigma>` have "\<not> SN_elt ?qr (r\<cdot>\<sigma>)" unfolding SN_elt_def by best
  then obtain t' where "r\<cdot>\<sigma> \<unrhd> t'" and "Tinf ?qr t'" using not_SN_imp_subt_Tinf by auto
  have "\<forall>x \<in> vars_term r. \<not>((Var x)\<cdot>\<sigma> \<unrhd> t')"
  proof
    fix x assume "x \<in> vars_term r"
    hence "SN_elt ?qr (Var x\<cdot>\<sigma>)" using xSN by simp
    show "\<not> Var x\<cdot>\<sigma> \<unrhd> t'"
    proof
      assume "Var x\<cdot>\<sigma> \<unrhd> t'"
      with `SN_elt ?qr (Var x\<cdot>\<sigma>)` have "SN_elt ?qr t'" using SN_rel_imp_SN_subt[OF ctxt_closed_qrstep] by auto
      with `Tinf ?qr t'` show "False" unfolding Tinf_def by auto
    qed
  qed
  with `r\<cdot>\<sigma> \<unrhd> t'` have "\<exists>u. r \<unrhd> u \<and> t' = u\<cdot>\<sigma>" using subt_instance_and_not_subst_imp_subt[where s = r and \<sigma> = \<sigma>] by auto
  then obtain u where "r \<unrhd> u" and "t' = u\<cdot>\<sigma>" by best
  have "\<forall>x. u \<noteq> Var x"
  proof (rule ccontr)
    assume "\<not> (\<forall>x. u \<noteq> Var x)"
    hence "\<exists>x. u = Var x" by simp
    then obtain x where "u = Var x" by best
    with `r \<unrhd> u` have "r \<unrhd> Var x" by auto
    hence "x \<in> vars_term r" using var_subt_imp_subt_in_vars by best
    with xSN `u = Var x` have "SN_elt ?qr (u\<cdot>\<sigma>)" by simp
    with `t' = u\<cdot>\<sigma>` have "SN_elt ?qr t'" by simp
    with `Tinf ?qr t'` show "False" unfolding Tinf_def by simp
  qed
  from `t' = u\<cdot>\<sigma>` and `Tinf ?qr t'` have "Tinf ?qr (u\<cdot>\<sigma>)" by simp
  have "\<not> l \<rhd> u"
  proof
    assume "l \<rhd> u"
    hence "l \<cdot> \<sigma> \<rhd> u \<cdot> \<sigma>" by (rule suptp_stable)
    with `Tinf ?qr (l\<cdot>\<sigma>)` have "SN_elt ?qr (u\<cdot>\<sigma>)" unfolding Tinf_def by best
    with `Tinf ?qr (u\<cdot>\<sigma>)` show "False" unfolding Tinf_def by auto
  qed
  from `r \<unrhd> u` have "r\<cdot>\<sigma> \<unrhd> (u\<cdot>\<sigma>)" by (rule supteqp_stable)
  from `(l,r) \<in> ?U` and `r \<unrhd> u` and `\<forall>x. u \<noteq> Var x` and `(?t,l\<cdot>\<sigma>) \<in> ?nrqr^*` 
    `(l\<cdot>\<sigma>,r\<cdot>\<sigma>) \<in> ?rqr` and `Tinf ?qr (u\<cdot>\<sigma>)` and `r\<cdot>\<sigma> \<unrhd> (u\<cdot>\<sigma>)` and `Tinf ?qr (l\<cdot>\<sigma>)`
    and `\<not> l \<rhd> u`
  show ?thesis by blast
qed

lemma SN_var:
 assumes "wwf_qtrs QR" shows "SN_elt (qrstep QR) (Var x)" (is "SN_elt _ ?x")
proof (cases QR)
  case (Pair Q R)
  let ?QR = "(Q,R)"
  let ?qr = "qrstep ?QR"
  have "SN_elt ?qr ?x"
  proof (rule ccontr)
    assume "\<not> ?thesis"
    then obtain A where "A 0 = ?x" and rsteps: "\<forall>i. (A i,A(Suc i)) \<in> ?qr"
      unfolding SN_elt_def by best
    hence "(?x,A 1) \<in> ?qr" using spec[OF rsteps, of 0] by auto
    then obtain l r C \<sigma> where "(l,r) \<in> R" and x: "Var x = C\<langle>l \<cdot> \<sigma>\<rangle>" unfolding qrstep_def rstep_r_c_s_def by auto
    with wwf_qtrs_imp_left_fun Pair assms obtain f ls where "l = Fun f ls" by blast
    with x obtain ls where "Var x = C\<langle>Fun f ls\<rangle>" by auto
    thus False by (cases C, auto)
  qed
  with Pair show ?thesis by auto
qed

lemma Tinf_var:
 assumes "wwf_qtrs QR" shows "\<not> Tinf (qrstep QR) (Var x)"
using SN_var[OF assms] unfolding Tinf_def by auto
 
lemma Tinf_lemma:
 assumes wwf: "wwf_qtrs QR" and tinf: "Tinf (qrstep QR) t"
 shows "\<exists>l r u \<sigma>. (l,r) \<in> useful_rules QR \<and> r \<unrhd> u \<and> (\<forall>x. u \<noteq> Var x) \<and> (t,l\<cdot>\<sigma>) \<in> (nr_qrstep QR)^* \<and> (l\<cdot>\<sigma>,r\<cdot>\<sigma>) \<in> r_qrstep QR
        \<and> r\<cdot>\<sigma> \<unrhd> (u\<cdot>\<sigma>) \<and> Tinf (qrstep QR) (l\<cdot>\<sigma>) \<and> Tinf (qrstep QR) (u\<cdot>\<sigma>) \<and> \<not>(l \<rhd> u)"
proof (cases QR)
  case (Pair Q R)
  show ?thesis 
  proof (cases t)
    case (Var x)
    with Tinf_var[OF wwf] tinf have False by auto
    thus ?thesis ..
  next
    case (Fun f ts)
    from Tinf_lemma_1[OF wwf[simplified Pair] tinf[simplified Fun Pair]]
    show ?thesis unfolding Pair Fun .
  qed
qed

lemma Tinf_imp_def_root_qrstep:
 assumes wwf: "wwf_qtrs (Q,R)" and tinf: "Tinf (qrstep (Q,R)) t"
 shows "\<exists>f ss. t = Fun f ss \<and> defined (useful_rules (Q,R)) f"
proof -
  from Tinf_var[OF wwf] tinf obtain f ts where t: "t = Fun f ts" by (cases t, auto)
  with tinf have tinf: "Tinf (qrstep (Q,R)) (Fun f ts)" by auto
  from Tinf_lemma[OF wwf tinf]
  obtain l r \<sigma> where lr: "(l,r) \<in> useful_rules (Q,R)" and "(Fun f ts,l\<cdot>\<sigma>) \<in> (nr_qrstep (Q,R))^*" 
    by blast
  with wwf[unfolded wwf_qtrs_wf_trs wf_trs_def]
  obtain g ls where l: "l = Fun g ls" by blast
  with nr_qrstep_seq_preserves_root[OF `(Fun f ts,l\<cdot>\<sigma>) \<in> (nr_qrstep (Q,R))^*`]
  have "g = f" by auto
  with lr l have "defined (useful_rules (Q,R)) f" unfolding defined_def by auto
  with t show ?thesis by auto
qed

definition wf_qtrs :: "('f,'v)qtrs \<Rightarrow> bool"
where "wf_qtrs QR \<equiv> wwf_qtrs QR \<and> \<not> (\<exists> x. Var x \<in> fst QR)"

lemma nr_qrstep_imp_sharp_qrstep: assumes step: "(s,t) \<in> nr_qrstep (Q,R)"
  shows "(sharp s, sharp t) \<in> qrstep (ID_QR (Q,R))"
proof -
  let ?QR = "(Q,R)"
  let ?qr = "qrstep ?QR"
  let ?nr = "nr_qrstep ?QR"
  let ?IQR = "ID_QR ?QR"
  let ?Iqr = "qrstep ?IQR"
  from step obtain l r C \<sigma> where C: "C \<noteq> \<box>" and  lr: "(l,r) \<in> R" and s: "s = C\<langle>l\<cdot>\<sigma>\<rangle>" and t: "t = C\<langle>r\<cdot>\<sigma>\<rangle>" and NF: "\<forall> u. l \<cdot> \<sigma> \<rhd> u \<longrightarrow>  NF_lhs Q u" 
    unfolding nr_qrstep_def rstep_r_c_s_def by auto
  from C obtain D f bef aft where C: "C = More f bef D aft" by (cases C, auto)
  let ?Dl = "D\<langle>l \<cdot> \<sigma>\<rangle>"
  let ?Dr = "D\<langle>r \<cdot> \<sigma>\<rangle>"
  let ?C = "More (Sharp f) (map plain bef) \<box> (map plain aft)"
  from s t C have s: "s = Fun f (bef @ ?Dl # aft)" and t: "t = Fun f (bef @ ?Dr # aft)" by auto
  from lr NF have "(?Dl,?Dr) \<in> ?qr" unfolding qrstep_def rstep_r_c_s_def by simp blast 
  with qrstep_iff_plain_qrstep have "(plain ?Dl, plain ?Dr) \<in> ?Iqr" by blast
  hence "(?C\<langle>plain ?Dl\<rangle>,?C\<langle>plain ?Dr\<rangle>) \<in> ?Iqr" by (rule qrstep_closed_ctxt )
  with s t show ?thesis by auto
qed

lemma nr_qrstep_seq_imp_sharp_qrstep_seq: assumes step: "(s,t) \<in> (nr_qrstep (Q,R))^*"
  shows "(sharp s, sharp t) \<in> (qrstep (ID_QR (Q,R)))^*"
using step
proof (induct rule: rtrancl_induct, simp)
  case (step u v)
  from nr_qrstep_imp_sharp_qrstep[OF step(2)] step(3)
  show ?case by auto
qed

lemma Tinf_lemma_DP_explicit_subst:
 assumes wf: "wf_qtrs QR" and s: "Tinf (qrstep QR) s"
 shows "\<exists>l r u \<sigma>. (l,r) \<in> DP (useful_rules QR) \<and> r\<cdot>\<sigma> = sharp u \<and> Tinf (qrstep QR) u \<and> (sharp s,l\<cdot>\<sigma>) \<in> (qrstep(ID_QR QR))^* \<and> NF_lhs (fst (ID_QR QR)) (l\<cdot>\<sigma>)"
proof (cases QR)
  case (Pair Q R)
  let ?QR = "(Q,R)"
  let ?qr = "qrstep ?QR"
  let ?U = "useful_rules ?QR"
  let ?nrqr = "nr_qrstep ?QR"
  let ?rqr = "r_qrstep ?QR"
  from wf Pair have wf: "wwf_qtrs ?QR" and wfQ: "\<not> (\<exists> x. Var x \<in> Q)" unfolding wf_qtrs_def by auto
  let ?IQR = "ID_QR ?QR"
  let ?Iqr = "qrstep ?IQR"
  obtain l r u \<sigma> where "(l,r) \<in> ?U" and "r \<unrhd> u" and "\<forall>x. u \<noteq> Var x" and "(s,l\<cdot>\<sigma>) \<in> ?nrqr^*"
    and step: "(l\<cdot>\<sigma>,r\<cdot>\<sigma>) \<in> ?rqr" and "Tinf ?qr (l\<cdot>\<sigma>)" and "Tinf ?qr (u\<cdot>\<sigma>)" and "\<not> l \<rhd> u"
    using Tinf_lemma[OF wf s[simplified Pair]] by auto
  let ?u = "u\<cdot>\<sigma>"
  let ?l = "sharp l"
  let ?r = "sharp u"
  let ?\<sigma> = "Subst(plain \<circ> (case \<sigma> of Subst s \<Rightarrow> s))"
  obtain f us where "u = Fun f us" by (cases u) (auto simp: `\<forall>x. u \<noteq> Var x`)
  have "root u = root ?u" unfolding `u = Fun f us` by simp
  obtain h vs where "?u = Fun h vs" and "defined ?U h" using Tinf_imp_def_root_qrstep[OF wf `Tinf ?qr ?u`] by auto
  from `(s,l\<cdot>\<sigma>) \<in> ?nrqr^*` have "(sharp s,sharp(l\<cdot>\<sigma>)) \<in> ?Iqr^*" by (rule nr_qrstep_seq_imp_sharp_qrstep_seq)
  from `(l,r) \<in> ?U` and wf[unfolded wwf_qtrs_wf_trs] obtain g ts where l: "l = Fun g ts" unfolding wf_trs_def by blast
  have "sharp(l\<cdot>\<sigma>) = ?l\<cdot>?\<sigma>" unfolding `l = Fun g ts` sharp_subst_distr by simp
  with `(sharp s,sharp(l\<cdot>\<sigma>)) \<in> ?Iqr^*` have "(sharp s,?l\<cdot>?\<sigma>) \<in> ?Iqr^*" by simp
  from `(u\<cdot>\<sigma>) = Fun h vs` and `u = Fun f us` have "f = h" by simp
  from `defined ?U h` have "defined ?U f" unfolding `f = h` by simp
  from `(l,r) \<in> ?U` and `r \<unrhd> u` and `defined ?U f` and `\<not>(l \<rhd> u)` have "(?l,?r) \<in> DP ?U" unfolding DP_def `u = Fun f us` by auto
  have "?r\<cdot>?\<sigma> = sharp ?u" unfolding `u = Fun f us` sharp_subst_distr by simp
  from step have NF: "\<forall> u. l\<cdot>\<sigma> \<rhd> u \<longrightarrow> NF_lhs Q u" unfolding r_qrstep_def rstep_r_c_s_def by auto
  have "NF_lhs (plain ` Q) (?l\<cdot>?\<sigma>)"
  proof (rule ccontr)
    assume "\<not> ?thesis"
    then obtain q C \<gamma> where q: "q \<in> plain ` Q" and match: "?l\<cdot>?\<sigma> = C\<langle>q \<cdot> \<gamma>\<rangle>" unfolding NF_lhs_def by auto
    from q obtain uq where uq: "q = plain uq" and uqQ: "uq \<in> Q" by auto
    let ?ug = "Subst (unsharp \<circ> (case \<gamma> of Subst s \<Rightarrow> s))"
    show False
    proof (cases C)
      case (More ff bef D aft)
      hence "sharp (Fun g ts)\<cdot>?\<sigma> \<rhd> q \<cdot> \<gamma>" using nectxt_imp_suptp[OF _ match] l by auto
      hence "sharp (Fun g ts \<cdot> \<sigma>) \<rhd> q \<cdot> \<gamma>" unfolding sharp_subst_distr by (cases \<sigma>, auto)
      hence "unsharp (sharp (l \<cdot> \<sigma>)) \<rhd> unsharp (q \<cdot> \<gamma>)" unfolding l by (rule  suptp_unsharp)
      hence "l \<cdot> \<sigma> \<rhd> unsharp (q \<cdot> \<gamma>)" by simp
      hence "l \<cdot> \<sigma> \<rhd> unsharp q \<cdot> ?ug" using unsharp_subst_distr by simp     
      with NF have NF: "NF_lhs Q (unsharp q \<cdot> ?ug)" by auto
      from q have us: "unsharp q \<in> Q" by auto
      have "\<exists> l C \<sigma>. l \<in> Q \<and> unsharp q \<cdot> ?ug = C\<langle>l \<cdot> \<sigma>\<rangle>" 
        by (rule exI[of _ "unsharp q"], rule exI[of _ \<box>], auto simp: us)
      with NF show False unfolding NF_lhs_def by blast
    next
      case Hole
      with match have match: "?l\<cdot>?\<sigma> = q \<cdot> \<gamma>" by auto
      show ?thesis 
      proof (cases uq)
        case (Fun ff qq)
        with l uq match show ?thesis by auto
      next
        case (Var x)
        with uqQ show False using wfQ by auto
      qed
    qed
  qed
  hence "NF_lhs (fst ?IQR) (?l\<cdot>?\<sigma>)" unfolding ID_QR_def by auto
  with `(?l,?r) \<in> DP ?U` and `?r\<cdot>?\<sigma> = sharp ?u` and `Tinf ?qr ?u` and `(sharp s,?l\<cdot>?\<sigma>) \<in> ?Iqr^*`
  show ?thesis unfolding Pair by blast
qed


fun next4 :: "('f,'v)qtrs \<Rightarrow> ('f,'v)term \<Rightarrow>
 (('f shp,'v)term \<times> ('f shp,'v)term \<times> ('f,'v)term \<times> ('f shp,'v)subst)"
where "next4 (Q,R) s = (SOME (l,r,u,\<sigma>).
 (l,r) \<in> DP (useful_rules (Q,R)) \<and> r\<cdot>\<sigma> = sharp u \<and> Tinf (qrstep (Q,R)) u \<and> (sharp s,l\<cdot>\<sigma>) \<in> (qrstep(ID_QR (Q,R)))^* \<and> NF_lhs (fst (ID_QR (Q,R))) (l\<cdot>\<sigma>))"

fun concrete_seq :: "('f,'v)qtrs \<Rightarrow> ('f,'v)term \<Rightarrow> nat \<Rightarrow>
 (('f,'v)term \<times> ('f shp,'v)term \<times> ('f shp,'v)term \<times> ('f,'v)term \<times> ('f shp,'v)subst)"
where "concrete_seq QR s 0       = (s,next4 QR s)"
    | "concrete_seq QR r (Suc n) = (let (_,_,_,s,_) = concrete_seq QR r n in (s,next4 QR s))"

lemma someI_ex4: "\<exists>w x y z. P(w,x,y,z) \<Longrightarrow> P(SOME(w,x,y,z). P(w,x,y,z))" by (simp add: some_eq_ex)


theorem not_SN_imp_min_ichain:
 fixes QR :: "('f,'v)qtrs"
 assumes "wf_qtrs QR" and "\<not> SN(qrstep QR)"
 shows "\<exists>s t \<sigma>. min_ichain (initial_dpp QR) s t \<sigma>"
proof (cases QR)
  case (Pair Q R)
  let ?QR = "(Q,R)"
  let ?qr = "qrstep ?QR"
  let ?IQR = "ID_QR ?QR"
  let ?Iqr = "qrstep ?IQR"
  from `\<not>(SN(qrstep QR))` Pair obtain s' :: "('f,'v)term" where "\<not>(SN_elt ?qr s')"
    unfolding SN_def by auto
  then obtain s where "s' \<unrhd>  s" and "Tinf ?qr s" using not_SN_imp_subt_Tinf by auto
  from assms Pair have "wf_qtrs ?QR" by auto
 let ?I = "\<lambda>i. fst(concrete_seq ?QR s i)"
 let ?S = "\<lambda>i. fst(snd(concrete_seq ?QR s i))"
 let ?T = "\<lambda>i. fst(snd(snd(concrete_seq ?QR s i)))"
 let ?\<sigma> = "\<lambda>i. snd(snd(snd(snd(concrete_seq ?QR s i))))"
 let ?P = "\<lambda>s l r u \<sigma>. (l,r) \<in> DP (useful_rules ?QR) \<and> r\<cdot>\<sigma> = sharp u \<and> Tinf ?qr u \<and> (sharp s,l\<cdot>\<sigma>) \<in> ?Iqr^* \<and> NF_lhs (fst ?IQR) (l\<cdot>\<sigma>)"
 let ?P' =  "\<lambda>(l,r,u,\<sigma>). ?P s l r u \<sigma>"
 have main: "\<forall>i. ?P (?I i) (?S i) (?T i) (?I(Suc i)) (?\<sigma> i)"
 proof
  fix i show "?P (?I i) (?S i) (?T i) (?I(Suc i)) (?\<sigma> i)"
  proof (induct i)
   case 0 
   have "\<exists>l r u \<sigma>. ?P'(l,r,u,\<sigma>)" using Tinf_lemma_DP_explicit_subst[OF `wf_qtrs ?QR` `Tinf ?qr s`] by auto
   hence "?P'(next4 ?QR s)" unfolding next4.simps using someI_ex4[where P = "?P'"] by best
   have ini: "?I 0 = s" unfolding concrete_seq.simps by simp
   have s: "?S 0 = fst(next4 ?QR s)" unfolding concrete_seq.simps by simp
   have t: "?T 0 = fst(snd(next4 ?QR s))" unfolding concrete_seq.simps by simp
   have sub: "?\<sigma> 0 = snd(snd(snd(next4 ?QR s)))" unfolding concrete_seq.simps Let_def split_def by auto
   have nxt: "?I(Suc 0) = fst(snd(snd(next4 ?QR s)))" unfolding concrete_seq.simps Let_def split_def by auto
   show ?case unfolding ini s t sub nxt using `?P'(next4 ?QR s)` by auto
  next
   case (Suc i)
   let ?P'' = "\<lambda>(w,x,y,z). ?P (?I(Suc i)) w x y z"
   from Suc have "Tinf ?qr (?I(Suc i))" by simp
   have "\<exists>l r u \<sigma>. ?P''(l,r,u,\<sigma>)" using Tinf_lemma_DP_explicit_subst[OF `wf_qtrs ?QR` `Tinf ?qr (?I(Suc i))`] by auto
   hence IH: "?P''(next4 ?QR (?I(Suc i)))" unfolding next4.simps using someI_ex4[where P = "?P''"] by simp
   have s: "?S(Suc i) = fst(next4 ?QR (?I(Suc i)))" unfolding concrete_seq.simps Let_def split_def by simp
   have t: "?T(Suc i) = fst(snd(next4 ?QR (?I(Suc i))))" unfolding concrete_seq.simps Let_def split_def by simp
   have sub: "?\<sigma>(Suc i) = snd(snd(snd(next4 ?QR (?I(Suc i)))))" unfolding concrete_seq.simps Let_def split_def by simp
   have nxt: "?I(Suc(Suc i)) = fst(snd(snd(next4 ?QR (?I(Suc i)))))" unfolding concrete_seq.simps Let_def split_def by simp
   show ?case unfolding s t sub nxt using IH by auto
  qed
 qed
 from main have P: "\<forall>i. (?S i,?T i) \<in> DP (useful_rules ?QR)" by simp
 from main have NF: "\<And>i. NF_lhs (fst ?IQR) (?S i \<cdot> ?\<sigma> i)" by auto
 from main have eq: "\<And>i. sharp(?I(Suc i)) = ?T i \<cdot> ?\<sigma> i" by auto
 have R: "\<forall>i. (?T i \<cdot> ?\<sigma> i,?S(Suc i) \<cdot> ?\<sigma>(Suc i)) \<in> ?Iqr^*"
 proof
  fix i show "(?T i \<cdot> ?\<sigma> i,?S(Suc i)\<cdot>?\<sigma>(Suc i)) \<in> ?Iqr^*"
  proof (cases i)
   case 0
   from main have first: "?T 0 \<cdot> ?\<sigma> 0 = sharp(?I(Suc 0))" by best
   from main have "(sharp(?I(Suc 0)),?S(Suc 0)\<cdot>?\<sigma>(Suc 0)) \<in> ?Iqr^*" by best
   thus ?thesis unfolding 0 unfolding first by simp
  next
   case (Suc i) show ?thesis unfolding eq[symmetric] using main by best
  qed
 qed
 from P and R and NF have ichain: "ichain (initial_dpp ?QR) ?S ?T ?\<sigma>" by auto
 have Tinf: "\<forall>i. Tinf ?qr (?I i)"
 proof
  fix i show "Tinf ?qr (?I i)" using main by (induct i) (auto simp: `Tinf ?qr s`)
 qed
 have SN_t: "\<forall>i. SN_elt ?Iqr (?T i \<cdot> ?\<sigma> i)"
 proof
  fix i
  from Tinf have "Tinf ?qr (?I(Suc i))" by best
  from eq have eq': "sharp(?I(Suc i)) = ?T i \<cdot> ?\<sigma> i" by simp
  from `wf_qtrs ?QR` have "wwf_qtrs ?QR" unfolding wf_qtrs_def by auto
  from Tinf_imp_SN_sharp[OF `wwf_qtrs ?QR` `Tinf ?qr (?I(Suc i))`]
  show "SN_elt ?Iqr (?T i \<cdot> ?\<sigma> i)" unfolding eq' by simp
 qed
 from ichain and SN_t
 show ?thesis unfolding initial_dpp.simps min_ichain.simps minimal_cond_def by (simp only: Pair) blast
qed


corollary not_SN_imp_min_ichain_rstep: assumes wf: "wf_trs R" and nSN: "\<not> SN (rstep R)" 
  shows "\<exists> s t \<sigma>. min_ichain (DP R, ({},ID R)) s t \<sigma>"
proof -
  let ?QR = "({},R)"
  let ?ID_QR = "ID_QR ?QR"
  let ?DP = "DP (useful_rules ?QR)"
  have U: "useful_rules ?QR = R" unfolding useful_rules_def useful_rule_def by auto
  from wf U have wf: "wf_qtrs ?QR" unfolding wf_qtrs_def wwf_qtrs_wf_trs by auto
  from nSN have nSN: "\<not> SN (qrstep ?QR)" unfolding qrstep_into_rstep_conv .
  have id: "initial_dpp ?QR = (DP R, ({}, ID R))" by (simp add: U ID_QR_def)
  from not_SN_imp_min_ichain[OF wf nSN, unfolded id] show ?thesis .
qed

hide_const next4 concrete_seq

end
