(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory DpTransImpl imports TermImpl DpFramework begin

types ('f,'v)dp = "('f shp,'v)rule" 
types ('f,'v)dpL = "('f,'v)dp list"


fun ID_list :: "('f,'v)rules \<Rightarrow> ('f shp,'v)rules"
where "ID_list R = map (\<lambda> lr. (plain (fst lr), plain (snd lr))) R"

lemma ID_list[simp]: "set (ID_list R) = ID (set R)"
proof (induct R, simp add: ID_def)
  case (Cons lr R)
  have "ID (set (lr # R)) = {(s,t). \<exists> l r. s = plain l \<and> t = plain r \<and> (l, r) \<in> set (lr # R)}" by (simp add: ID_def)
  also have "\<dots> = {(s,t). \<exists> l r. s = plain l \<and> t = plain r \<and> (l, r) \<in> set R} \<union> {(s,t). \<exists> l r. s = plain l \<and> t = plain r \<and> (l, r) = lr}" by auto
  also have "\<dots> = set (ID_list R) \<union> {(s,t). \<exists> l r. s = plain l \<and> t = plain r \<and> (l, r) = lr}" using Cons by (auto simp: ID_def)
  finally show ?case by (cases lr, auto)
qed

declare ID_list.simps[simp del]

definition
  check_wf_trs :: "('f::show, 'v::show) rules \<Rightarrow> shows check"
where
  "check_wf_trs R \<equiv> (
     check_allm (\<lambda>rule. do {
       check (\<not> is_var (fst rule)) (shows_string ''variable left-hand side in'');
       check_subseteq (vars_term_list (snd rule)) (vars_term_list (fst rule))
         <+? (\<lambda>x. ''free variable '' +#+ shows x
               +@+ shows_string '' in right-hand side of'')
     } <+? (\<lambda>s. s +@+ '' rule '' +#+ shows_rule rule +@+ shows_nl)
     ) R
   <+? (\<lambda>e. ''the TRS is not well-formed'' +#+ shows_nl +@+ e)
   )"

lemma check_wf_trs_conf[simp]:
  shows "isOK (check_wf_trs R) = wf_trs (set R)"
proof
  assume "isOK (check_wf_trs R)"
  hence "\<forall>(l, r)\<in>set R. \<not> is_var l" unfolding check_wf_trs_def by auto
  hence "\<forall>(l, r)\<in>set R. \<exists>f ts. l = Fun f ts" using not_is_var_imp_Fun by best
  moreover have "\<forall>(l, r)\<in>set R. vars_term r \<subseteq> vars_term l"
  proof (rule ballI2)
    fix l r assume "(l, r) \<in> set R"
    from `isOK (check_wf_trs R)`
      have "isOK(check_allm (\<lambda>(l, r). check_subseteq (vars_term_list r) (vars_term_list l)) R)"
      unfolding check_wf_trs_def by auto
    hence "\<forall>(l, r)\<in>set R. isOK (check_subseteq (vars_term_list r) (vars_term_list l))"
      unfolding split_def by simp
    with `(l, r) \<in> set R` show "vars_term r \<subseteq> vars_term l" by auto
  qed
  ultimately show "wf_trs (set R)" unfolding wf_trs_def by auto
next
  assume "wf_trs (set R)" thus "isOK (check_wf_trs R)"
    unfolding wf_trs_def check_wf_trs_def by force
qed

definition DP_list :: "('f,'v)rules \<Rightarrow> ('f shp,'v)rules"
where "DP_list R \<equiv> let D = defs_list R in concat (map (\<lambda> lr. let l = fst lr; s = sharp l in map (\<lambda> u. (s, sharp u)) (filter (\<lambda> u. \<not> l \<rhd> u \<and> \<not> is_var u \<and> the (root u) mem D) (supteq_list (snd lr)))) R)"

lemma DP_list[simp]: "set (DP_list R) = DP (set R)"
proof -
  let ?p = "\<lambda> s t l r u. s = sharp l \<and> t = sharp u \<and> (l,r) \<in> set R \<and> \<not> l \<rhd> u \<and> r \<unrhd> u \<and> \<not> is_var u \<and> defined (set R) (the (root u))"
  let ?q = "\<lambda> s t l r h us.  s = sharp l \<and> t = sharp (Fun h us) \<and> (l, r) \<in> set R \<and> r \<unrhd> Fun h us \<and> defined (set R) h \<and> \<not> l \<rhd> Fun h us"
  have "set (DP_list R) = (\<Union>(l,r) \<in> set R. ((\<lambda>u. (sharp l, sharp u)) ` {u. r \<unrhd> u \<and> (l \<unrhd> u \<longrightarrow> l = u) \<and> \<not> is_var u \<and> defined (set R) (the (root u))}))"
    unfolding DP_list_def by (auto simp: Let_def defs_list_sound[symmetric])
  also have "\<dots> = {(s,t). \<exists> l r u. ?p s t l r u}" by auto
  also have "\<dots> = {(s,t). \<exists> l r h us. ?q s t l r h us}" 
  proof -
    {
      fix s t l r
      have "(\<exists> u. ?p s t l r u) = (\<exists> h us. ?q s t l r h us)" (is "?l = ?r")
      proof
        assume ?l
        then obtain u where "?p s t l r u" ..
        then obtain h us where "?q s t l r h us" by (cases u, auto)
        thus ?r by blast
      next
        assume ?r 
        then obtain h us where "?q s t l r h us" by auto
        hence "?p s t l r (Fun h us)" by auto
        thus ?l by blast
      qed          
    }
    thus ?thesis by simp
  qed
  also have "\<dots> = DP (set R)" unfolding DP_def by simp
  finally show ?thesis .
qed
      
definition
  check_DP :: "('f::show,'v::show)rules \<Rightarrow> ('f,'v)dpL \<Rightarrow> shows check"
where
  "check_DP r P \<equiv> do {
     check_wf_trs r;
     check_subseteq (DP_list r) P 
       <+? (\<lambda>dp. ''the DP '' +#+ shows_rule dp
         +@+ '' does not appear in the DP problem'' +#+ shows_nl)
   } <+? (\<lambda>s. ''the DP-transformation is not applied correctly.'' +#+ shows_nl +@+ s)"

lemma check_DP:
  assumes check: "isOK(check_DP r P)"
  shows "finite_dpp (set P, ID (set r)) \<Longrightarrow> SN (rstep (set r))"
proof -
  assume fin: "finite_dpp (set P, ID (set r))"
  show "SN (rstep (set r))"
  proof -
    from check have wf: "wf_trs (set r)" 
      and dp: "Trs.DP (set r) \<subseteq> set P" 
      unfolding check_DP_def by auto
    let ?DP = "Trs.DP (set r)"
    from fin have "finite_dpp (set P, ID (set r))"  by simp
    hence fin2: "finite_dpp (set P \<inter> ?DP, ID (set r))" by (unfold finite_dpp_def, auto)
    from dp have "?DP = set P \<inter> ?DP" by auto
    with fin2 have "finite_dpp (DP (set r), ID (set r))" by simp
    with wf show "?thesis" by (rule finite_initial_dpp_imp_SN)
  qed
qed

definition
  check_not_wf_trs :: "('f::show, 'v::show) rules \<Rightarrow> shows check"
where
  "check_not_wf_trs R \<equiv>
     check (\<not> isOK (check_wf_trs R)) (''The TRS is well formed'' +#+ shows_nl)"

lemma check_not_wf_trs:
  assumes check: "isOK(check_not_wf_trs R)"
  shows "\<not> SN (rstep (set R))"
proof -
  from assms have "\<not> wf_trs (set R)" unfolding check_not_wf_trs_def by auto
  with SN_rstep_imp_wf_trs show ?thesis by auto
qed

definition
  check_no_var :: "('f::show, 'v::show) term \<Rightarrow> shows check"
where
  "check_no_var t \<equiv> check (\<not> is_var t) (''variable found'' +#+ shows_nl)"

lemma check_no_var_sound[simp]:
  assumes "isOK (check_no_var t)" shows "\<not> is_var t"
using assms unfolding check_no_var_def by simp

definition
  check_no_defined_root :: "'f::show list \<Rightarrow> ('f, 'v::show) term \<Rightarrow> shows check"
where
  "check_no_defined_root fs t = check (\<not> the (root t) mem fs) (
     ''the root of '' +#+ shows t +@+ '' is defined w.r.t. the signature''
     +#+ shows_nl +@+ shows fs +@+ shows_nl
   )"

lemma check_no_defined_root_sound[simp]:
  assumes "isOK (check_no_defined_root (defs_list R) t)"
  shows "\<not> defined (set R) (the (root t))"
using assms unfolding check_no_defined_root_def by (simp add: mem_iff defs_list_sound)

end
