(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Innermost imports Trs begin

definition NF_lhs :: "('f,'v)term set \<Rightarrow> ('f,'v)term \<Rightarrow> bool"
where "NF_lhs ls t \<equiv> \<not> (\<exists> l C \<sigma>. l \<in> ls \<and> t = C\<langle>l \<cdot> \<sigma>\<rangle>)"

lemma NF_lhs_NF_rstep_conv: "NF_lhs (fst ` R) t = (t \<in> NF (rstep R))" (is "?l = ?r")
proof
  assume ?l
  show ?r
  proof
    fix s
    show "(t,s) \<notin> rstep R" 
    proof
      assume "(t,s) \<in> rstep R" 
      with rstep_iff_rstep_r_c_s[of t s R] obtain l r C \<sigma> where lr: "(l,r) \<in> R" and step: "(t,s) \<in> rstep_r_c_s (l,r) C \<sigma>" by blast
      from step have t: "t = C\<langle>l \<cdot> \<sigma>\<rangle>" unfolding rstep_r_c_s_def by auto
      with `?l` lr show False unfolding NF_lhs_def by force
    qed
  qed
next
  assume ?r
  show ?l unfolding NF_lhs_def
  proof 
    assume "\<exists> l C \<sigma>. l \<in> fst ` R \<and> t = C\<langle>l \<cdot> \<sigma>\<rangle>"
    then obtain l C \<sigma> where l: "l \<in> fst ` R" and t:"t = C\<langle>l \<cdot> \<sigma>\<rangle>" by auto
    from l obtain r where lr: "(l,r) \<in> R" by auto
    from t have "(t,C\<langle>r \<cdot> \<sigma>\<rangle>) \<in> rstep_r_c_s (l,r) C \<sigma>" unfolding rstep_r_c_s_def by auto
    with lr have "(t,C\<langle>r \<cdot> \<sigma>\<rangle>) \<in> rstep R" unfolding rstep_iff_rstep_r_c_s by blast
    with NF_no_step[OF `?r`] show False by auto
  qed
qed

lemma not_NF_lhs_lhs: assumes l: "l \<in> Q" shows "\<not> NF_lhs Q l"
  by (unfold NF_lhs_def, simp, rule exI[of _ l], simp add: l, rule exI[of _ \<box>], rule exI[of _ \<iota>], simp)

lemma NF_lhs_instance: assumes nf: "NF_lhs ls (t \<cdot> \<sigma>)"
  shows "NF_lhs ls t"
proof (rule ccontr)  
  assume "\<not> ?thesis"
  then obtain l C \<sigma>' where l: "l \<in> ls" and t: "t = C\<langle>l \<cdot> \<sigma>'\<rangle>" unfolding NF_lhs_def by auto
  have "t \<cdot> \<sigma> = (C \<cdot>c \<sigma>)\<langle>l \<cdot> (\<sigma>' \<circ>s \<sigma>)\<rangle>" by (simp add: t)
  thus False using nf l unfolding NF_lhs_def by blast
qed


lemma NF_lhs_subterm: assumes nf: "NF_lhs ls t" and sub: "t \<unrhd> s"
  shows "NF_lhs ls s"
proof (rule ccontr)  
  assume "\<not> ?thesis"
  then obtain l C \<sigma> where l: "l \<in> ls" and s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" unfolding NF_lhs_def by auto
  obtain D where "t = D\<langle>s\<rangle>" using supteqp_imp_ctxt[OF sub] by auto
  with s have "t = (D \<circ>c C)\<langle>l \<cdot> \<sigma>\<rangle>" by auto
  thus False using nf l unfolding NF_lhs_def by blast
qed


definition NF_lhs_set :: "('f,'v)term set \<Rightarrow> ('f,'v)term set"
where "NF_lhs_set Q \<equiv> {t | t. NF_lhs Q t}"

types ('f,'v)qtrs = "('f,'v)term set \<times> ('f,'v)trs"

definition NF_Q_R_check :: "('f,'v)qtrs \<Rightarrow> bool"
where "NF_Q_R_check QR \<equiv> (\<forall> l \<in> fst ` (snd QR). (\<exists> q C \<sigma>. q \<in> fst QR \<and> l = C\<langle>q \<cdot> \<sigma>\<rangle>))"

lemma NF_Q_R_check: "NF_Q_R_check (Q,R) = (NF_lhs_set Q \<subseteq> NF (rstep R))" (is "?l = ?r")
proof 
  assume ?r
  show ?l unfolding NF_Q_R_check_def
  proof
    fix l
    assume "l \<in> fst ` snd (Q,R)"
    hence l: "l \<in> fst ` R" by simp
    have "l \<notin> NF (rstep R)" 
      by (simp only: NF_lhs_NF_rstep_conv[symmetric], unfold NF_lhs_def, simp, rule exI[of _ l], simp add: l,
        rule exI[of _ Hole], rule exI[of _ \<iota>], auto)
    with `?r` have "\<not> NF_lhs Q l" unfolding NF_lhs_set_def by auto
    thus "\<exists>q C \<sigma>. q \<in> (fst (Q,R)) \<and> l = C\<langle>q \<cdot> \<sigma>\<rangle>" unfolding NF_lhs_def by auto
  qed
next
  assume ?l
  show ?r
  proof
    fix t
    assume nf: "t \<in> NF_lhs_set Q"
    show "t \<in> NF (rstep R)"
    proof (rule ccontr)
      assume "\<not> ?thesis"
      then obtain l C \<sigma> where lR: "l \<in> fst ` R" and t: "t = C\<langle>l \<cdot> \<sigma>\<rangle>"
        by (simp only: NF_lhs_NF_rstep_conv[symmetric], unfold NF_lhs_def, auto)
      from `?l` lR obtain q C' \<sigma>' where q: "q \<in> Q" and l: "l = C'\<langle>q \<cdot> \<sigma>'\<rangle>" unfolding NF_Q_R_check_def by auto
      have "t = C\<langle>C'\<langle>q \<cdot> \<sigma>'\<rangle> \<cdot> \<sigma>\<rangle>" by (simp add: t l)
      also have "\<dots> = C\<langle>(C' \<cdot>c \<sigma>)\<langle>q \<cdot> (\<sigma>' \<circ>s \<sigma>)\<rangle>\<rangle>" by simp
      also have "\<dots> = (C \<circ>c (C' \<cdot>c \<sigma>))\<langle>q \<cdot> (\<sigma>' \<circ>s \<sigma>)\<rangle>" by simp
      finally show False using nf q unfolding NF_lhs_set_def NF_lhs_def by blast
    qed
  qed
qed

lemma NF_lhs_mono: "R \<subseteq> R' \<Longrightarrow> NF_lhs R' t \<Longrightarrow> NF_lhs R t" unfolding NF_lhs_def by auto
lemma NF_lhs_empty: "NF_lhs {} t" unfolding NF_lhs_def by auto

(* Q-restricted rewriting *)
definition qrstep :: "('f,'v)qtrs \<Rightarrow> ('f,'v)trs"
where "qrstep QR = {(s,t) | s t. (\<exists> l r C \<sigma>. (l,r) \<in> snd QR \<and> (s,t) \<in> rstep_r_c_s (l,r) C \<sigma> \<and> (\<forall> u. l \<cdot> \<sigma> \<rhd> u \<longrightarrow> NF_lhs (fst QR) u))}"

(* with the special case innermost rewriting *)
definition irstep :: "('f,'v)trs \<Rightarrow> ('f,'v)trs"
where "irstep R \<equiv> qrstep (fst ` R, R)"

(* and the special case standard rewriting *)
lemma qrstep_into_rstep_conv: "qrstep ({},R) = rstep R"
unfolding qrstep_def
using rstep_iff_rstep_r_c_s
unfolding rstep_CC_CS
by (simp add: NF_lhs_empty, blast)

lemma qrstep_mono_r: "R \<subseteq> R' \<Longrightarrow> qrstep (Q,R) \<subseteq> qrstep (Q,R')"
unfolding qrstep_def by auto

lemma qrstep_mono_q: "Q \<subseteq> Q' \<Longrightarrow> qrstep (Q',R) \<subseteq> qrstep (Q,R)"
using NF_lhs_mono[of Q Q'] unfolding qrstep_def by simp blast

lemma qrstep_subset_rstep: "qrstep (Q,R) \<subseteq> rstep R"
  by (simp only: qrstep_into_rstep_conv[symmetric], rule qrstep_mono_q, auto)

lemma q_NF_imp_rstep_NF: assumes QR: "NF_Q_R_check (Q,R)"
  and nf: "NF_lhs Q t" shows "t \<in> NF (rstep R)"
using NF_Q_R_check assms unfolding NF_lhs_set_def by auto


lemma q_NF_imp_qrstep_NF: assumes QR: "NF_Q_R_check (Q,R)"
  and nf: "NF_lhs Q t" shows "t \<in> NF (qrstep (Q,R))"
using q_NF_imp_rstep_NF[OF QR nf] NF_anti_mono[OF qrstep_subset_rstep] by blast

lemma NF_lhs_suptp_is_NF_lhs_args: "(\<forall> u. Fun f ls \<rhd> u \<longrightarrow> NF_lhs Q u) = (\<forall> l. l \<in> set ls \<longrightarrow> NF_lhs Q l)" (is "?l = ?r")
proof 
  assume ?l
  show ?r
  proof (intro impI allI)
    fix l
    assume "l \<in> set ls"
    hence "Fun f ls \<rhd> l" by auto
    with `?l` show "NF_lhs Q l" by auto
  qed
next
  assume ?r
  show ?l
  proof (intro impI allI)
    fix u
    assume supt: "Fun f ls \<rhd> u"
    from suptp_imp_nectxt[OF supt] obtain C where "C \<noteq> \<box>" and "Fun f ls = C\<langle>u\<rangle>" by auto
    then obtain b C a where "Fun f ls = Fun f (b @ C\<langle>u\<rangle> # a)" by (cases C, auto)
    then obtain l where "l \<in> set ls" and "l = C\<langle>u\<rangle>" by auto
    with `?r` have "NF_lhs Q C\<langle>u\<rangle>" by auto
    with NF_lhs_subterm[OF _ ctxt_imp_supteqp[of C u]]
    show "NF_lhs Q u" by simp
  qed
qed



lemma rstep_r_c_s_closed_ctxt: "((C\<langle>s\<rangle>,C\<langle>t\<rangle>) \<in> rstep_r_c_s lr (C \<circ>c D) \<sigma>) = ((s,t) \<in> rstep_r_c_s lr D \<sigma>)"
unfolding rstep_r_c_s_def by auto

lemma qrstep_closed_ctxt: assumes "(s,t) \<in> qrstep QR"
  shows "(C\<langle>s\<rangle>,C\<langle>t\<rangle>) \<in> qrstep QR"
proof (cases QR)
  case (Pair Q R)
  with assms obtain l r D \<sigma> where lr: "(l,r) \<in> R" and step: "(s,t) \<in> rstep_r_c_s (l,r) D \<sigma>" and nf: "\<forall> u. l \<cdot> \<sigma> \<rhd> u \<longrightarrow> NF_lhs Q u" 
    unfolding qrstep_def by auto
  from step have "(C\<langle>s\<rangle>,C\<langle>t\<rangle>) \<in> rstep_r_c_s (l,r) (C \<circ>c D) \<sigma>" by (simp add: rstep_r_c_s_closed_ctxt)
  with lr nf Pair show ?thesis unfolding qrstep_def by auto
qed
    
lemma ctxt_closed_qrstep: "ctxt_closed (qrstep QR)"
unfolding ctxt_closed_def
proof(rule subsetI2)
  fix s t
  assume "(s,t) \<in> CC (qrstep QR)"
  thus "(s,t) \<in> qrstep QR"
    by (induct, auto simp: qrstep_closed_ctxt)
qed

lemma NF_irstep_is_NF_rstep: "NF (irstep R) = NF (rstep R)" (is "?l = ?r")
proof 
  show "?r \<subseteq> ?l"
    by (rule NF_anti_mono, unfold irstep_def, simp add: qrstep_into_rstep_conv[symmetric], rule qrstep_mono_q, auto)
next
  show "?l \<subseteq> ?r"
  proof
    fix t
    assume nfi: "t \<in> NF (irstep R)"
    show "t \<in> NF (rstep R)"
    proof(intro NF_I allI) 
      fix s
      obtain S where S: "S = {u | u. (\<exists> D. t = D\<langle>u\<rangle>) \<and> (\<exists> l C \<sigma>. l \<in> fst ` R \<and> u = C\<langle>l \<cdot> \<sigma>\<rangle>)}" by auto
      have tD: "\<exists> D. t = D\<langle>t\<rangle>" by (rule exI[of _ \<box>], simp) 
      show "(t,s) \<notin> rstep R"
      proof (rule ccontr)
        assume "\<not> ?thesis"
        hence tS: "t \<in> S" using rstep_iff_rstep_r_c_s[of t s R] tD unfolding rstep_r_c_s_def S by force
        from mp[OF spec[OF spec[OF SN_imp_minimal[OF SN_supt]]] tS]
        obtain u where u: "u \<in> S" and min: "\<forall> u'. (u,u') \<in> supt \<longrightarrow> u' \<notin> S" by blast
        from u obtain D l C \<sigma> where tu: "t = D\<langle>u\<rangle>" and l: "l \<in> fst ` R" and C: "u = C\<langle>l \<cdot> \<sigma>\<rangle>" unfolding S by force
        from l obtain r where lr: "(l,r) \<in> R" by auto
        let ?s = "(D \<circ>c C)\<langle>r \<cdot> \<sigma>\<rangle>"
        from tu C have "t = (D \<circ>c C)\<langle>l \<cdot> \<sigma>\<rangle>" by auto
        with lr have step: "(t,?s) \<in> rstep_r_c_s (l,r) (D \<circ>c C) \<sigma>" unfolding rstep_r_c_s_def by auto
        with spec[OF NF_no_step[OF nfi], of ?s] lr have "\<not> (\<forall> u . l \<cdot> \<sigma> \<rhd> u \<longrightarrow> NF_lhs (fst ` R) u)"
          unfolding irstep_def qrstep_def by simp
        then obtain u' where lsu': "l \<cdot> \<sigma> \<rhd> u'" and u'redex: "\<not> (NF_lhs (fst ` R) u')" by auto
        from u'redex have redex: "\<exists> l C \<sigma>. l \<in> fst ` R \<and> u' = C\<langle>l \<cdot> \<sigma>\<rangle>" unfolding NF_lhs_def by auto
        from suptp_imp_nectxt[OF lsu'] obtain E where E: "E \<noteq> \<box>" and El: "l \<cdot> \<sigma> = E\<langle>u'\<rangle>" by auto
        with C have "u = (C \<circ>c E)\<langle>u'\<rangle> \<and> C \<circ>c E \<noteq> \<box>" by (cases C, auto)
        then obtain E where E: "u = E\<langle>u'\<rangle> \<and> E \<noteq> \<box>" by blast
        hence uu': "u \<rhd> u'" using nectxt_imp_suptp by blast
        hence supt: "(u,u') \<in> supt" unfolding supt_def by auto
        have tu': "\<exists> D. t = D\<langle>u'\<rangle>" by (rule exI[of _ "D \<circ>c E"], simp add: tu E)
        with supt redex min show False unfolding S by blast
      qed
    qed
  qed
qed

definition useful_rule :: "('f,'v)term set \<Rightarrow> ('f,'v)rule \<Rightarrow> bool"
where "useful_rule Q lr \<equiv> \<not> (\<exists> s. fst lr \<rhd> s \<and> (\<exists> q \<in> Q. \<exists> \<sigma>. s =  q \<cdot> \<sigma>))"

lemma useful_rule_empty: "useful_rule {} lr" 
  unfolding useful_rule_def by auto

lemma only_useful_rules: assumes "\<forall> u. l \<cdot> \<sigma> \<rhd> u \<longrightarrow> NF_lhs Q u"
  shows "useful_rule Q (l,r)"
proof (rule ccontr)
  assume "\<not> ?thesis"
  then obtain s q \<delta> where q: "q \<in> Q" and supt: "l \<rhd> s" and match: "s = q \<cdot> \<delta>" unfolding useful_rule_def by auto
  with mp[OF spec[OF assms] suptp_stable[OF supt]] have "NF_lhs Q (q \<cdot> (\<delta> \<circ>s \<sigma>))" by auto
  hence "q \<cdot> (\<delta> \<circ>s \<sigma>) \<noteq> \<box>\<langle>q \<cdot> (\<delta> \<circ>s \<sigma>)\<rangle>" using q unfolding NF_lhs_def by blast
  thus False by simp
qed

definition useful_rules :: "('f,'v)qtrs \<Rightarrow> ('f,'v)trs"
where "useful_rules QR \<equiv> { lr | lr. lr \<in> snd QR \<and> useful_rule (fst QR) lr }"

lemma useful_rules_empty: "useful_rules ({},R) = R" unfolding useful_rules_def using useful_rule_empty by auto

lemma qrstep_useful_rules: "qrstep (Q, useful_rules (Q,R)) = qrstep (Q,R)"
proof -
  let ?QR = "(Q,R)"
  let ?uQR = "useful_rules ?QR"
  let ?QuQR = "(Q, ?uQR)"
  show ?thesis
  proof
    show "qrstep ?QuQR \<subseteq> qrstep ?QR"
      by (rule qrstep_mono_r, unfold useful_rules_def, auto)
  next
    show "qrstep ?QR \<subseteq> qrstep ?QuQR"
    proof (rule subsetI2)
      fix s t
      assume "(s,t) \<in> qrstep ?QR"
      then obtain l r C \<sigma> where lr: "(l,r) \<in> R" and step: "(s,t) \<in> rstep_r_c_s (l,r) C \<sigma>" and NF: "\<forall> u. l \<cdot> \<sigma> \<rhd> u \<longrightarrow> NF_lhs Q u" unfolding qrstep_def by auto
      from only_useful_rules[OF NF] have "useful_rule Q (l,r)" .
      with lr have "(l,r) \<in> ?uQR" unfolding useful_rules_def by auto
      with step NF show "(s,t) \<in> qrstep ?QuQR" unfolding qrstep_def by auto
    qed
  qed
qed      

definition wwf_rule :: "('f,'v)term set \<Rightarrow> ('f,'v)rule \<Rightarrow> bool"
where "wwf_rule Q lr \<equiv> case lr of Pair l r \<Rightarrow> useful_rule Q lr \<longrightarrow> ((\<exists> f ts. l = Fun f ts) \<and> vars_term l \<supseteq> vars_term r)"

definition wwf_qtrs :: "('f,'v)qtrs \<Rightarrow> bool"
where "wwf_qtrs QR \<equiv> \<forall> lr \<in> snd QR. wwf_rule (fst QR) lr"

lemma wwf_qtrs_wf_trs: "wwf_qtrs (Q,R) = wf_trs (useful_rules (Q,R))"
unfolding wwf_qtrs_def wwf_rule_def wf_trs_def useful_rules_def by force  

lemma wwf_qtrs_empty: "wwf_qtrs ({},R) = wf_trs R"
  unfolding wwf_qtrs_wf_trs useful_rules_empty by simp

lemma left_var_imp_not_SN_qrstep: fixes R :: "('f,'v)trs" and t :: "('f,'v)term"
  assumes lr: "(Var y,r) \<in> R" (is "(?y,_) \<in> _") shows 
  "\<not> (SN_elt (qrstep (Q,R)) t)"
proof -
  have steps: "\<forall> t. \<exists> s. (t,s) \<in> qrstep (Q,R)"
  proof 
    fix t :: "('f,'v)term"
    show "\<exists> s. (t,s) \<in> qrstep (Q,R)"
    proof (induct t)
      case (Var x)
      let ?yt = "subst y (Var x)"
      let ?ryt = "r \<cdot> ?yt"
      show ?case
        by (rule exI[of _ ?ryt], unfold qrstep_def, simp, intro exI, rule conjI, rule lr, rule exI[of _ \<box>], rule exI[of _ ?yt],
        unfold rstep_r_c_s_def, simp add: subst_def, simp only: suptp_iff_supteqp_neq[symmetric], simp only: suptp_var, auto)
    next
      case (Fun f ss)
      show ?case 
      proof (cases ss)
        case Nil
        let ?yt = "subst y (Fun f ss)"
        let ?ryt = "r \<cdot> ?yt"
        show ?thesis
          by (rule exI[of _ ?ryt], unfold qrstep_def, simp, intro exI, rule conjI, rule lr, rule exI[of _ \<box>], rule exI[of _ ?yt],
            unfold rstep_r_c_s_def, simp add: subst_def, simp only: suptp_iff_supteqp_neq[symmetric], simp only: Nil suptp_const, simp)
      next
        case (Cons s ts)
        with Fun obtain t where step: "(s,t) \<in> qrstep (Q,R)" by auto 
        let ?C = "More f [] \<box> ts"
        from Cons have id: "Fun f ss = ?C\<langle>s\<rangle>" by auto
        have "(?C\<langle>s\<rangle>, ?C\<langle>t\<rangle>) \<in> CC (qrstep (Q,R))"
           by (rule ctxt_closureI, rule step)
        with ctxt_closed_qrstep[of "(Q,R)"]  
        have "(Fun f ss, ?C\<langle>t\<rangle>) \<in> qrstep (Q,R)"
          unfolding ctxt_closed_def id by auto
        thus ?thesis by auto
      qed
    qed
  qed
  show ?thesis 
  proof
    assume "SN_elt (qrstep (Q,R)) t"
    from this obtain s where norm: "(t,s) \<in> (qrstep (Q,R))^!" using SN_elt_imp_normalizability by force
    from steps obtain u where "(s,u) \<in> qrstep (Q,R)" by auto
    with norm show False by auto
  qed
qed        

(* TODO: replace in Trs.thy *)

lemma not_SN_subt_imp_not_SN:
 assumes ctxt: "ctxt_closed R" and SN: "\<not> SN_elt R t" and sub: "s \<unrhd> t"
 shows "\<not> SN_elt R s"
using SN_rel_imp_SN_subt[OF ctxt _ sub] SN by auto


lemma SN_elt_imp_wwf_rule: assumes SN: "SN_elt (qrstep (Q,R)) t"
  and t: "t = C\<langle>l \<cdot> \<sigma>\<rangle>" and lr: "(l,r) \<in> R" and NF: "\<forall> u. l \<cdot> \<sigma> \<rhd> u \<longrightarrow> NF_lhs Q u"
  shows "wwf_rule Q (l,r)"
proof (cases "vars_term l \<supseteq> vars_term r")
  case True
  from left_var_imp_not_SN_qrstep[of _ _ R Q t] lr SN obtain f ls where l: "l = Fun f ls" by (cases l, auto)
  with True l show ?thesis unfolding wwf_rule_def by auto
next
  case False
  then obtain x where x: "x \<in> vars_term r - vars_term l" by auto
  hence xn: "x \<notin> vars_term l" by auto
  from only_useful_rules[OF NF] have ulr: "useful_rule Q (l,r)" .
  obtain \<delta> where delta: "\<delta> = Subst (\<lambda> y. if y = x then l \<cdot> \<sigma> else case \<sigma> of Subst s \<Rightarrow> s y)" by auto
  have "l \<cdot> \<sigma> = l \<cdot> (\<sigma>|_(vars_term l))" by (rule coincidence_lemma)
  also have "\<dots> = l \<cdot> (\<delta>|_(vars_term l))"  
  proof (rule arg_cong[of _ _ "op \<cdot> l "])
    show "\<sigma>|_vars_term l = \<delta>|_vars_term l" 
    proof (cases \<sigma>)
      case (Subst sig)
      {
        fix y
        have " (if y \<in> vars_term l then sig y else Var y) = (if y \<in> vars_term l then if y = x then l \<cdot> \<sigma> else sig y else Var y)" by (cases "y = x", auto simp: xn)
      } note main = this
      show ?thesis by (simp add: Subst delta, rule ext, simp add: main Subst)
    qed
  qed
  also have "\<dots> = l \<cdot> \<delta>" by (rule coincidence_lemma[symmetric])
  finally have lsld: "l \<cdot> \<sigma> = l \<cdot> \<delta>" .
  from x have xr: "x \<in> vars_term r" by auto
  from supteqp_imp_ctxt[OF supteqp_var[OF xr]] obtain C where "r = C\<langle>Var x\<rangle>" by auto
  hence "r \<cdot> \<delta> = (C \<cdot>c \<delta>)\<langle>l \<cdot> \<delta>\<rangle>" by (auto simp: lsld delta)
  then obtain C where rd:  "r \<cdot> \<delta> = C\<langle>l \<cdot> \<delta>\<rangle>" by auto
  obtain ls where ls: "l \<cdot> \<sigma> = ls" by auto
  obtain f where f: "\<And> i. f i = (C^i)\<langle>l \<cdot> \<delta>\<rangle>" by auto
  have steps: "\<forall> i. (f i, f (Suc i)) \<in> qrstep (Q,R)"
  proof
    fix i
    show "(f i, f (Suc i)) \<in> qrstep (Q,R)"
    proof (simp only: f, unfold qrstep_def, simp, intro exI, rule conjI, rule lr, rule exI[of _ "C^i"], rule exI[of _ \<delta>], simp add: rstep_r_c_s_def rd lsld[symmetric] NF[simplified] ls times_ctxt_def)
      have "C\<langle>(C^i)\<langle>ls\<rangle>\<rangle> = (C^(1 + i))\<langle>ls\<rangle>" by (simp only: ctxt_power_compose_distr, simp add: one_ctxt_def times_ctxt_def) 
      also have "\<dots> = (C^(i + 1))\<langle>ls\<rangle>" by auto
      also have "\<dots> = (C ^ i)\<langle>C\<langle>ls\<rangle>\<rangle>" by (simp only: ctxt_power_compose_distr, simp add: one_ctxt_def times_ctxt_def) 
      finally show "C\<langle>(C ^ i)\<langle>ls\<rangle>\<rangle> = (C ^ i)\<langle>C\<langle>ls\<rangle>\<rangle>" .
    qed
  qed
  have start: "f 0 = l \<cdot> \<sigma>" 
    by (simp add: f lsld[symmetric] ls one_ctxt_def)
  from start steps have nSN: "\<not> (SN_elt (qrstep (Q,R)) (l \<cdot> \<sigma>))" unfolding SN_elt_def by auto
  from not_SN_subt_imp_not_SN[OF ctxt_closed_qrstep nSN ctxt_imp_supteqp] SN[simplified t]
  have False ..
  thus ?thesis ..
qed

lemma SN_imp_wwf_qtrs: assumes "SN (qrstep QR)" shows "wwf_qtrs QR"
proof (cases QR)
  case (Pair Q R)
  let ?QR = "(Q,R)"
  let ?qr = "qrstep ?QR"
  {
    assume "\<not> wwf_qtrs ?QR"
    then obtain l r where lr: "(l,r) \<in> R" and nwwf: "\<not> wwf_rule Q (l,r)" unfolding wwf_qtrs_def by auto
    from nwwf have usef: "useful_rule Q (l,r)" unfolding wwf_rule_def by auto
    from assms Pair have SN: "SN_elt ?qr l" unfolding SN_defs by auto
    from nwwf SN_elt_imp_wwf_rule[OF SN _ lr, of \<box> \<iota>] have "\<not> (\<forall> u. l \<rhd> u \<longrightarrow> NF_lhs Q u)" by auto
    then obtain u where supt: "l \<rhd> u" and nNF: "\<not> NF_lhs Q u" by auto
    from nNF obtain C q \<sigma> where q: "q \<in> Q" and u: "u = C\<langle>q \<cdot> \<sigma>\<rangle>" unfolding NF_lhs_def by auto
    let ?qs = "q \<cdot> \<sigma>"
    from suptp_imp_nectxt[OF supt] obtain D where D: "D \<noteq> \<box>" and "l = D\<langle>u\<rangle>" by auto
    with u have l: "l = (D \<circ>c C)\<langle>?qs\<rangle>" by auto
    from D have DC: "D \<circ>c C \<noteq> \<box>" by (cases D, auto)
    with usef[unfolded useful_rule_def] nectxt_imp_suptp[OF DC l] q
    have False by auto
  }
  thus ?thesis using Pair by auto
qed

lemma left_var_useful: "useful_rule Q (Var x,r)"
unfolding useful_rule_def using suptp_var[of x] by auto

lemma wwf_qtrs_imp_left_fun: assumes "wwf_qtrs (Q,R)" and "(l,r) \<in> R" shows "\<exists> f ls. l = Fun f ls"
using assms unfolding wwf_qtrs_def wwf_rule_def using left_var_useful[of Q _ r] by (cases l, auto)
    

definition r_qrstep :: "('f,'v)qtrs \<Rightarrow> ('f,'v)trs"
where "r_qrstep QR = {(s,t) | s t. (\<exists> l r \<sigma>. (l,r) \<in> snd QR \<and> (s,t) \<in> rstep_r_c_s (l,r) \<box> \<sigma> \<and> (\<forall> u. l \<cdot> \<sigma> \<rhd> u \<longrightarrow> NF_lhs (fst QR) u))}"

definition nr_qrstep :: "('f,'v)qtrs \<Rightarrow> ('f,'v)trs"
where "nr_qrstep QR = {(s,t) | s t. (\<exists> l r \<sigma> C. (l,r) \<in> snd QR \<and> C \<noteq> \<box> \<and> (s,t) \<in> rstep_r_c_s (l,r) C \<sigma> \<and> (\<forall> u. l \<cdot> \<sigma> \<rhd> u \<longrightarrow> NF_lhs (fst QR) u))}"

lemma qrstep_iff_r_qrstep_or_nr_qrstep: "qrstep QR = r_qrstep QR \<union> nr_qrstep QR" (is "?l = ?r \<union> ?nr")
proof -
  have C: "\<And> C. C = \<box> \<or> C \<noteq> \<box>" by auto
  have "?nr \<subseteq> ?l" unfolding qrstep_def nr_qrstep_def by auto
  moreover have "?r \<subseteq> ?l" unfolding qrstep_def r_qrstep_def by auto
  moreover have "?l \<subseteq> ?r \<union> ?nr" 
  proof (rule subsetI2)
    fix s t
    assume "(s,t) \<in> qrstep QR"
    then obtain l r \<sigma> C where "(l,r) \<in> snd QR \<and> (s,t) \<in> rstep_r_c_s (l,r) C \<sigma> \<and> (\<forall> u. l \<cdot> \<sigma> \<rhd> u \<longrightarrow> NF_lhs (fst QR) u)" unfolding qrstep_def by auto
    thus "(s,t) \<in> ?r \<union> ?nr" unfolding r_qrstep_def nr_qrstep_def
      by (cases "C = \<box>", blast+)
  qed
  ultimately show ?thesis by blast
qed

lemma nr_qrstep_preserves_root: assumes "(Fun f ss, t) \<in> nr_qrstep QR" 
  shows "\<exists> ts. t = Fun f ts"
proof -
  from assms obtain C l r \<sigma> where t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>" and "C \<noteq> \<box>" and "Fun f ss = C\<langle>l \<cdot> \<sigma>\<rangle>" unfolding nr_qrstep_def rstep_r_c_s_def by auto
  then obtain bef D aft where "C = More f bef D aft" by (cases C, auto)
  with t have "t = Fun f (bef @ (D\<langle>r \<cdot> \<sigma>\<rangle> # aft))" by auto
  thus ?thesis ..
qed

lemma nr_qrstep_seq_preserves_root:
 assumes "(Fun f ss,t) \<in> (nr_qrstep QR)^*"
 shows "\<exists>ts. t = (Fun f ts)"
using assms by induct (auto simp: nr_qrstep_preserves_root)

lemma nr_qrstep_equiv_num_args: assumes step: "(s,t) \<in> nr_qrstep (Q,R)" shows "num_args s = num_args t"
proof -
  let ?QR = "(Q,R)"
  let ?nr = "nr_qrstep ?QR"
  from step obtain l r C \<sigma> where C: "C \<noteq> \<box>" and  lr: "(l,r) \<in> R" and s: "s = C\<langle>l\<cdot>\<sigma>\<rangle>" and t: "t = C\<langle>r\<cdot>\<sigma>\<rangle>" and NF: "\<forall> u. l \<cdot> \<sigma> \<rhd> u \<longrightarrow>  NF_lhs Q u" 
    unfolding nr_qrstep_def rstep_r_c_s_def by auto
  from C obtain D f bef aft where C: "C = More f bef D aft" by (cases C, auto)
  from s t C show ?thesis by auto
qed

lemma nr_qrstep_seq_equiv_num_args: assumes step: "(s,t) \<in> (nr_qrstep (Q,R))^*" shows "num_args s = num_args t"
  using assms by (induct rule: rtrancl.induct) (auto simp: nr_qrstep_equiv_num_args)

lemma nr_qrstep_imp_arg_qrstep: assumes step: "(s,t) \<in> nr_qrstep (Q,R)"
  and i: "i < num_args s" shows "(args s ! i, args t ! i) \<in> (qrstep (Q,R))^="
proof -
  let ?QR = "(Q,R)"
  let ?qr = "qrstep ?QR"
  let ?nr = "nr_qrstep ?QR"
  from step obtain l r C \<sigma> where C: "C \<noteq> \<box>" and  lr: "(l,r) \<in> R" and s: "s = C\<langle>l\<cdot>\<sigma>\<rangle>" and t: "t = C\<langle>r\<cdot>\<sigma>\<rangle>" and NF: "\<forall> u. l \<cdot> \<sigma> \<rhd> u \<longrightarrow>  NF_lhs Q u" 
    unfolding nr_qrstep_def rstep_r_c_s_def by auto
  from C obtain D f bef aft where C: "C = More f bef D aft" by (cases C, auto)
  let ?Dl = "D\<langle>l \<cdot> \<sigma>\<rangle>"
  let ?Dr = "D\<langle>r \<cdot> \<sigma>\<rangle>"
  let ?l = "length bef"
  let ?as = "args s ! i"
  let ?at = "args t ! i"
  from s t C have s: "s = Fun f (bef @ ?Dl # aft)" and t: "t = Fun f (bef @ ?Dr # aft)" by auto
  show ?thesis 
  proof (cases "i = ?l")
    case True
    hence "?as = ?Dl" and "?at = ?Dr" using s t by auto
    with lr NF have "(?as,?at) \<in> ?qr" unfolding qrstep_def rstep_r_c_s_def by auto
    thus ?thesis by auto
  next
    case False
    have "?as = ?at" 
    proof (cases "i < ?l", simp add: s t nth_append)
      case False
      with `i \<noteq> ?l` have "i - ?l > 0" by auto
      then obtain k where "i - ?l = Suc k" by (cases "i - ?l", auto)
      with False show ?thesis using s t by (simp add: nth_append)
    qed
    thus ?thesis by auto
  qed
qed

lemma nr_qrstep_seq_imp_arg_qrstep_seq: assumes step: "(s,t) \<in> (nr_qrstep (Q,R))^*"
  and i: "i < num_args s" shows "(args s ! i, args t ! i) \<in> (qrstep (Q,R))^*"
  using step 
proof (induct rule: rtrancl_induct, simp)
  case (step u v)
  from nr_qrstep_seq_equiv_num_args[OF step(1)] i
  have i: "i < num_args u" by auto
  from nr_qrstep_imp_arg_qrstep[OF step(2) i] step(3) show ?case by auto
qed

end
