(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory RelativeRewriting
imports Trs RedPair
begin

types ('f,'v)reltrs = "('f,'v)trs \<times> ('f,'v)trs"

types ('f,'v)reltrsL = "('f,'v)rules \<times> ('f,'v)rules"

fun rel_step :: "('f,'v)reltrs \<Rightarrow> ('f,'v)term ars"
where "rel_step (R,S) = (rstep S)^* O (rstep R) O (rstep S)^*"

lemma one_step_ctxt_closed_rel_step : "one_step_ctxt_closed (rel_step RS)"
proof (cases RS)
  case (Pair R S)
  from rtrancl_one_step_ctxt_closed[OF one_step_ctxt_closed_rstep[of S]] have S: "one_step_ctxt_closed ((rstep S)^*)" .
  from Pair comp_one_step_ctxt_closed[OF comp_one_step_ctxt_closed[OF S one_step_ctxt_closed_rstep[of R] ] S]
  show ?thesis by simp
qed
  
lemma subst_closed_rel_step : "subst_closed (rel_step RS)"
proof (cases RS)
  case (Pair R S)
  from subst_closed_rtrancl[OF subst_closed_rstep[of S]] have S: "subst_closed ((rstep S)^*)" .
  from Pair comp_subst_closed[OF S comp_subst_closed[OF subst_closed_rstep[of R] S] ]
  show ?thesis by simp
qed
  

definition rel_SN :: "('f,'v)reltrs \<Rightarrow> bool"
where "rel_SN RS \<equiv> SN (rel_step RS)"

fun rel_SN_alt :: "('f,'v)reltrs \<Rightarrow> bool"
where "rel_SN_alt (R,S) = (\<forall> (f :: nat \<Rightarrow> ('f,'v)term). (\<forall> i. (f i, f (Suc i)) \<in> rstep (R \<union> S)) \<longrightarrow> (\<exists> i. \<forall> j \<ge> i. (f j, f (Suc j)) \<notin> rstep R))"

lemma rel_SN_to_rel_SN_alt : "rel_SN (R,S) \<Longrightarrow> rel_SN_alt (R,S)"
proof (unfold rel_SN_def)
  assume SN: "SN (rel_step (R,S))"
  let ?R = "rstep R"
  let ?S = "rstep S"
  show ?thesis
  proof (simp, intro allI impI)
    fix f
    assume steps: "\<forall> i. (f i, f (Suc i)) \<in> rstep (R \<union> S)"
    hence steps: "\<And> i. (f i, f (Suc i)) \<in> ?R \<union> ?S" unfolding rstep_union by blast
    obtain r where  r: "\<And> j. r j \<equiv>  (f j, f (Suc j)) \<in> ?R" by auto
    show "\<exists> i. \<forall> j \<ge> i. (f j, f (Suc j)) \<notin> ?R"
    proof (rule ccontr)
      assume "\<not> ?thesis"
      hence ih: "infinite_hits r" unfolding infinite_hits_def r by blast
      obtain r_index where "r_index = infinite_hits.index r" by simp
      with infinite_hits.index_p[OF ih] infinite_hits.index_ordered[OF ih] infinite_hits.index_not_p_between[OF ih] 
      have r_index: "\<And> i. r (r_index i) \<and> r_index i < r_index (Suc i) \<and> (\<forall> j. r_index i < j \<and> j < r_index (Suc i) \<longrightarrow> \<not> r j)" by auto
      obtain g where g: "\<And> i. g i \<equiv> f (r_index i)" ..
      {
        fix i
        let ?ri = "r_index i"
        let ?rsi = "r_index (Suc i)"
        from r_index have isi: "?ri < ?rsi" by auto
        obtain ri rsi where ri: "ri = ?ri" and rsi: "rsi = ?rsi" by auto
        with r_index[of i] steps have inter: "\<And> j. ri < j \<and> j < rsi \<Longrightarrow> (f j, f (Suc j)) \<in> ?S" unfolding r by auto
        from ri isi rsi have risi: "ri < rsi" by simp                      
        {
          fix n
          assume "Suc n \<le> rsi - ri"
          hence "(f (Suc ri), f (Suc (n + ri))) \<in> ?S^*"
          proof (induct n, simp)
            case (Suc n)
            hence stepps: "(f (Suc ri), f (Suc (n+ri))) \<in> ?S^*" by simp
            have "(f (Suc (n+ri)), f (Suc (Suc n + ri))) \<in> ?S"
              using inter[of "Suc n + ri"] Suc(2) by auto
            with stepps show ?case by simp
          qed
        }
        from this[of "rsi - ri - 1"] risi have 
          "(f (Suc ri), f rsi) \<in> ?S^*" by simp
        with ri rsi have ssteps: "(f (Suc ?ri), f ?rsi) \<in> ?S^*" by simp
        with r_index[of i] have "(f ?ri, f ?rsi) \<in> ?R O ?S^*" unfolding r by auto
        hence "(g i, g (Suc i)) \<in> ?S^* O ?R O ?S^*" using rtrancl_refl unfolding g by auto           
      } 
      hence "\<not> SN (?S^* O ?R O ?S^*)" unfolding SN_defs by blast
      with SN show False unfolding rel_step.simps by simp
    qed
  qed
qed


lemma rtrancl_list_conv:  "((s,t) \<in> R^*) = 
  (\<exists> list.  last (s # list) = t \<and> (\<forall> i. i < length list \<longrightarrow> ((s # list) ! i, (s # list) ! Suc i) \<in> R))" (is "?l = ?r")
proof 
  assume ?r
  then obtain list where "last (s # list) = t \<and> (\<forall> i. i < length list \<longrightarrow> ((s # list) ! i, (s # list) ! Suc i) \<in> R)" ..
  thus ?l
  proof (induct list arbitrary: s, simp)
    case (Cons u ll)
    hence "last (u # ll) = t \<and> (\<forall> i. i < length ll \<longrightarrow> ((u # ll) ! i, (u # ll) ! Suc i) \<in> R)" by auto
    from Cons(1)[OF this] have rec: "(u,t) \<in> R^*" .
    from Cons have "(s, u) \<in> R" by auto
    with rec show ?case by auto
  qed
next
  assume ?l
  from rtrancl_imp_seq[OF this]
  obtain S n where s: "S 0 = s" and t: "S n = t" and steps: "\<forall> i<n. (S i, S (Suc i)) \<in> R" by auto
  let ?list = "map (\<lambda> i. S (Suc i)) [0 ..< n]"
  show ?r
  proof (rule exI[of _ ?list], intro conjI, 
      cases n, simp add: s[symmetric] t[symmetric], simp add: t[symmetric]) 
    show "\<forall> i < length ?list. ((s # ?list) ! i, (s # ?list) ! Suc i) \<in> R" 
    proof (intro allI impI)
      fix i
      assume i: "i < length ?list"
      thus "((s # ?list) ! i, (s # ?list) ! Suc i) \<in> R"
      proof (cases i, simp add: s[symmetric] steps)
        case (Suc j)
        with i steps show ?thesis by simp
      qed
    qed
  qed
qed


fun choice :: "(nat \<Rightarrow> 'a list) \<Rightarrow> nat \<Rightarrow> (nat \<times> nat)"
where "choice f 0 = (0,0)"
  |   "choice f (Suc n) = (let (i,j) = choice f n in 
           if Suc j < length (f i) 
               then (i,Suc j)
               else (Suc i, 0))"
        
lemma rel_SN_alt_to_rel_SN : "rel_SN_alt (R,S) \<Longrightarrow> rel_SN (R,S)"
proof (unfold rel_SN_def)
  assume SN: "rel_SN_alt (R,S)"
  let ?R = "rstep R"
  let ?S = "rstep S"
  show "SN (rel_step (R,S))"
  proof
    fix f
    assume  "\<forall> i. (f i, f (Suc i)) \<in> rel_step (R,S)"
    hence steps: "\<And> i. (f i, f (Suc i)) \<in> ?S^* O ?R O ?S^*" by auto
    let ?prop = "\<lambda> i ai bi. (f i, bi) \<in> ?S^* \<and> (bi, ai) \<in> ?R \<and> (ai, f (Suc (i))) \<in> ?S^*"
    {
      fix i
      from steps obtain bi ai where "?prop i ai bi" by blast
      hence "\<exists> ai bi. ?prop i ai bi" by blast
    }
    hence "\<forall> i. \<exists> bi ai. ?prop i ai bi" by blast
    from choice[OF this] obtain b where "\<forall> i. \<exists> ai. ?prop i ai (b i)" by blast
    from choice[OF this] obtain a where steps: "\<And> i. ?prop i (a i) (b i)" by blast
    let ?prop = "\<lambda> i li. (b i, a i) \<in> ?R \<and> (\<forall> j < length li. ((a i # li) ! j, (a i # li) ! Suc j) \<in> ?S) \<and> last (a i # li) = b (Suc i)"
    {
      fix i
      from steps[of i] steps[of "Suc i"] have "(a i, f (Suc i)) \<in> ?S^*" and "(f (Suc i), b (Suc i)) \<in> ?S^*" by auto
      from rtrancl_trans[OF this] steps[of i] have R: "(b i, a i) \<in> ?R" and S: "(a i, b (Suc i)) \<in> ?S^*" by blast+
      from S[unfolded rtrancl_list_conv] obtain li where "last (a i # li) = b (Suc i) \<and> (\<forall> j < length li. ((a i # li) ! j, (a i # li) ! Suc j) \<in> ?S)" ..
      with R have "?prop i li" by blast
      hence "\<exists> li. ?prop i li" ..
    }
    hence "\<forall> i. \<exists> li. ?prop i li" ..
    from choice[OF this] obtain l where steps: "\<And> i. ?prop i (l i)" by auto
    let ?p = "\<lambda> i. ?prop i (l i)"
    from steps have steps: "\<And> i. ?p i" by blast
    let ?l = "\<lambda> i. a i # l i"
    let ?g = "\<lambda> i. choice (\<lambda> j. ?l j) i"
    obtain g where g: "\<And> i. g i = (let (ii,jj) = ?g i in ?l ii ! jj)" by auto
    have len: "\<And> i j n. ?g n = (i,j) \<Longrightarrow> j < length (?l i)"
    proof -
      fix i j n
      assume n: "?g n = (i,j)"
      show "j < length (?l i)" 
      proof (cases n)
        case 0
        with n have "j = 0" by auto
        thus ?thesis by simp
      next
        case (Suc nn)
        obtain ii jj where nn: "?g nn = (ii,jj)" by (cases "?g nn", auto)
        show ?thesis 
        proof (cases "Suc jj < length (?l ii)")
          case True
          with nn Suc have "?g n = (ii, Suc jj)" by auto
          with n True show ?thesis by simp
        next
          case False 
          with nn Suc have "?g n = (Suc ii, 0)" by auto
          with n show ?thesis by simp
        qed
      qed
    qed      
    have gsteps: "\<And> i. (g i, g (Suc i)) \<in> ?R \<union> ?S"
    proof -
      fix n
      obtain i j where n: "?g n = (i, j)" by (cases "?g n", auto)
      show "(g n, g (Suc n)) \<in> ?R \<union> ?S"
      proof (cases "Suc j < length (?l i)")
        case True
        with n have "?g (Suc n) = (i, Suc j)" by auto
        with n have gn: "g n = ?l i ! j" and gsn: "g (Suc n) = ?l i ! (Suc j)" unfolding g by auto
        thus ?thesis using steps[of i] True by auto
      next
        case False
        with n have "?g (Suc n) = (Suc i, 0)" by auto
        with n have gn: "g n = ?l i ! j" and gsn: "g (Suc n) = a (Suc i)" unfolding g by auto
        from gn len[OF n] False have "j = length (?l i) - 1" by auto
        with gn have gn: "g n = last (?l i)" using last_conv_nth[of "?l i"] by auto
        from gn gsn show ?thesis using steps[of i] steps[of "Suc i"] by auto
      qed
    qed
    have infR:  "\<forall> n. \<exists> j \<ge> n. (g j, g (Suc j)) \<in> ?R" 
    proof
      fix n
      obtain i j where n: "?g n = (i,j)" by (cases "?g n", auto)
      from len[OF n] have j: "j \<le> length (?l i) - 1" by simp
      let ?k = "length (?l i) - 1 - j"
      obtain k where k: "k = j + ?k" by auto
      from j k have k2: "k = length (?l i) - 1" and k3: "j + ?k < length (?l i)" by auto
      {
        fix n i j k l
        assume n: "choice l n = (i,j)" and "j + k < length (l i)"
        hence "choice l (n + k) = (i, j + k)"
          by (induct k arbitrary: j, simp, auto)
      }
      from this[OF n, of ?k, OF k3]
      have gnk: "?g (n + ?k) = (i, k)" by (simp only: k)
      hence "g (n + ?k) = ?l i ! k" unfolding g by auto
      hence gnk2: "g (n + ?k) = last (?l i)" using last_conv_nth[of "?l i"] k2 by auto
      from k2 gnk have "?g (Suc (n+?k)) = (Suc i, 0)" by auto
      hence gnsk2: "g (Suc (n+?k)) = a (Suc i)" unfolding g by auto
      from steps[of i] steps[of "Suc i"] have main: "(g (n+?k), g (Suc (n+?k))) \<in> ?R" 
        by (simp only: gnk2 gnsk2)
      show "\<exists> j \<ge> n. (g j, g (Suc j)) \<in> ?R" 
        by (rule exI[of _ "n + ?k"], auto simp: main[simplified])
    qed      
    from gsteps have steps: "\<And> i. (g i, g (Suc i)) \<in> rstep (R \<union> S)" unfolding rstep_union .
    from SN[simplified] steps infR show False by blast 
  qed
qed

hide_const choice

lemma rel_SN_conv : "rel_SN = rel_SN_alt"
  by (intro ext, clarify, intro iffI, rule rel_SN_to_rel_SN_alt, simp, rule rel_SN_alt_to_rel_SN, simp)

definition sound_rel_tt_impl :: "(('f,'v)reltrsL \<Rightarrow> ('f,'v)reltrsL \<Rightarrow> bool) \<Rightarrow> bool"
where "sound_rel_tt_impl tt \<equiv> \<forall> R S rR rS. (tt (R,S) (rR,rS) \<longrightarrow> rel_SN_alt (set rR,set rS) \<longrightarrow> rel_SN_alt (set R,set S))"

definition sound_rel_tt :: "(('f,'v)reltrs \<Rightarrow> ('f,'v)reltrs \<Rightarrow> bool) \<Rightarrow> bool"
where "sound_rel_tt tt \<equiv> \<forall> RS rRS. (tt RS rRS \<longrightarrow> rel_SN_alt rRS \<longrightarrow> rel_SN_alt RS)"

lemma rel_SN_r_empty : "rel_SN_alt ({}, S)"
unfolding rel_SN_alt.simps by auto

lemma rel_SN_s_empty : "rel_SN_alt (R, {}) = SN (rstep R)"
unfolding rel_SN_alt.simps SN_defs by auto

lemma apply_sound_rel_tt_impl : assumes sound: "sound_rel_tt_impl tt"
  and tt: "tt (R,S) (rR,rS)"
  and relSN: "rel_SN_alt (set rR, set rS)"
  shows "rel_SN_alt (set R, set S)"
using assms unfolding sound_rel_tt_impl_def by auto
      
  

context redpair
begin

lemma relativeRuleRemoval:
  assumes monotonic: "one_step_ctxt_closed \<succ>"
  and nonStrict: "r \<union> s \<subseteq> \<succeq> \<union> \<succ>"
  and strict: "(r - rr) \<union> (s - rs) \<subseteq> \<succ>"
  and rel_SN: "rel_SN_alt (rr,rs)"
  shows "rel_SN_alt (r,s)"
proof (simp, intro allI impI)
  fix t
  assume alli: "\<forall>i. (t i, t (Suc i)) \<in> rstep (r \<union> s)"
  show "\<exists> i. \<forall> j \<ge> i. (t j, t (Suc j)) \<notin> rstep r"
  proof (cases "\<forall> i. \<exists> j \<ge> i. (t j, t (Suc j)) \<in> (rstep r - rstep rr) \<union> (rstep s - rstep rs)")
    case True
    from monotonic strict substS trs_inclusion have "(rstep (r - rr)) \<subseteq> \<succ>" by auto
    with rstep_diff have one: "rstep r - rstep rr \<subseteq> \<succ>" by auto
    from monotonic strict substS trs_inclusion have "(rstep (s - rs)) \<subseteq> \<succ>" by auto
    with rstep_diff have two: "rstep s - rstep rs \<subseteq> \<succ>" by auto
    from one two True alli have strict: "\<forall> i. \<exists> j \<ge> i. (t j, t (Suc j)) \<in> \<succ>" by blast 
    from trs_inclusion[OF union_one_step_ctxt_closed[OF wmNS monotonic] subst_closed_union[OF substNS substS] nonStrict]
      alli have "\<forall>i. (t i, t (Suc i)) \<in> (\<succeq> \<union> \<succ>)" by auto
    from non_strict_ending[OF this compat] SN obtain j where "\<forall> i \<ge> j. (t i, t (Suc i)) \<in> \<succeq> - \<succ>" unfolding SN_def by auto
    with strict have False by auto
    thus ?thesis ..
  next    
    case False
    then obtain i where not: "\<forall> j \<ge> i. (t j, t (Suc j)) \<notin> (rstep r - rstep rr) \<union> (rstep s - rstep rs)" by blast
    with alli[unfolded rstep_union] have "\<forall> j \<ge> i. (t j, t (Suc j)) \<in> rstep (rr \<union> rs)" unfolding rstep_union by blast
    hence "\<forall> j. (shift t i j, shift t i (Suc j)) \<in> rstep (rr \<union> rs)" by auto
    with spec[OF rel_SN[simplified], of "shift t i"] obtain j where not2: "\<forall> k \<ge> j. (shift t i k, shift t i (Suc k)) \<notin> rstep rr" by blast
    show ?thesis
    proof (rule exI[of _ "i + j"], intro allI impI)
      fix l
      assume l: "i + j \<le> l"
      then obtain k where k: "k = l - i" and kj: "k \<ge> j"  by auto
      from k l have shift: "(t l, t (Suc l)) = (shift t i k, shift t i (Suc k))" by simp
      from not l have not: "(t l, t (Suc l)) \<notin> (rstep r - rstep rr)" by auto
      from not2 kj have not2: "(t l, t (Suc l)) \<notin> rstep rr" by (simp add: shift)
      from not not2 show "(t l, t (Suc l)) \<notin> rstep r" by auto
    qed
  qed
qed
end

fun
  check_relative_rule_removal ::
    "('f :: show, 'v :: show)redpair_impl \<Rightarrow> 
     ('f,'v)reltrsL \<Rightarrow> ('f,'v)reltrsL \<Rightarrow> shows check"
where
  "check_relative_rule_removal rp (R,S) (rR,rS)  = do {
     let (Sr :: ('f,'v) rules , NSr) = List.partition (\<lambda> foo. isOK(rp_strict rp foo)) R;
     let (Ss :: ('f,'v) rules , NSs) = List.partition (\<lambda> foo. isOK(rp_strict rp foo)) S;
     rp_valid rp;
     rp_mono rp
       <+? (\<lambda>s. ''problem with monotonicity of strict order'' +#+ shows_nl +@+ s);
     check_allm (rp_non_strict rp) NSr
       <+? (\<lambda>s. ''problem when orienting rules'' +#+ shows_nl +@+ s);
     check_allm (rp_non_strict rp) NSs
       <+? (\<lambda>s. ''problem when orienting rules'' +#+ shows_nl +@+ s);
     check_subseteq NSs rS
       <+? (\<lambda>s. case (rp_strict rp) s of Inl err \<Rightarrow> err | Inr ok \<Rightarrow> internal_error);
     check_subseteq NSr rR
       <+? (\<lambda>s. case (rp_strict rp) s of Inl err \<Rightarrow> err | Inr ok \<Rightarrow> internal_error)
    } <+? (\<lambda>s. ''could not apply the rule removal processor with the following'' +#+ shows_nl +@+
         rp_desc rp +@+ shows_nl +@+ s)"

context generic_redpair_impl
begin

lemma relative_rule_removal_sound : 
  shows "sound_rel_tt_impl (\<lambda> RS rRS. isOK(check_relative_rule_removal (generate_redpair rp) RS rRS))"
unfolding sound_rel_tt_impl_def
proof (intro allI impI)
  let ?rp = "generate_redpair rp"
  let ?af = "rp_af ?rp"
  fix R S rR rS
  assume check: "isOK (check_relative_rule_removal ?rp (R,S) (rR,rS))"
    and relSN: "rel_SN_alt (set rR, set rS)"
  show "rel_SN_alt (set R, set S)"
  proof (cases "List.partition (\<lambda> r. isOK(rp_strict ?rp r)) R")
    case (Pair Sr NSr) note rPair = this
    let ?NSr = "set NSr \<inter> set rR"
    show ?thesis
    proof (cases "List.partition (\<lambda> r. isOK(rp_strict ?rp r)) S") 
      case (Pair Ss NSs) 
      let ?NSs = "set NSs \<inter> set rS"
      from Pair rPair check have 
        one: "isOK(check_allm (rp_strict ?rp) (Ss @ Sr))" and two: "isOK(check_allm (rp_non_strict ?rp) (NSs @ NSr))" 
        and mono: "isOK(rp_mono ?rp)" and valid: "isOK(rp_valid ?rp)"  by (auto simp: Let_def mem_iff) 
      from generate_redpair[OF valid one two] mono obtain s NS where redpair: "mono_ce_af_redpair s NS ?af" and Sr: "set Sr \<subseteq> s" and NSr: "set NSr \<subseteq> NS" and Ss: "set Ss \<subseteq> s" and NSs: "set NSs \<subseteq> NS" and mono: "one_step_ctxt_closed s"
        by auto
      from rPair Sr NSr check have "set R - set NSr \<subseteq> s \<and> ?NSr \<inter> set R \<subseteq> NS \<union> s \<and> set NSr \<subseteq> set rR" by (auto)
      with check have NSr: "set R \<subseteq> NS \<union> s" and Sr: "set R - set rR \<subseteq> s" by auto
      from Pair Ss NSs check have "set S - set NSs \<subseteq> s \<and> ?NSs \<inter> set S \<subseteq> NS \<union> s \<and> set NSs \<subseteq> set rS" by auto
      with check have NSs: "set S \<subseteq> NS \<union> s" and Ss: "set S - set rS \<subseteq> s"  by auto
      from redpair interpret mono_ce_af_redpair s NS ?af .
      show ?thesis by (rule relativeRuleRemoval[OF mono _ _ relSN], auto simp: NSr NSs Sr Ss)
    qed
  qed
qed
end

lemma rel_SN_alt_mono: assumes R: "R \<subseteq> R'" and S: "S \<subseteq> S'"
  and SN: "rel_SN_alt (R',S')"
  shows "rel_SN_alt (R,S)"
proof (unfold rel_SN_alt.simps, intro allI impI)
  fix f
  assume "\<forall> i. (f i, f (Suc i)) \<in> rstep (R \<union> S)"
  with rstep_mono[OF R] rstep_mono[OF S] have
    "\<forall> i. (f i, f (Suc i)) \<in> rstep (R' \<union> S')"
    by (simp only: rstep_union, auto)
  with SN obtain i where "\<forall> j \<ge> i. (f j, f (Suc j)) \<notin> rstep R'" by auto
  thus "\<exists> i. \<forall> j \<ge> i. (f j, f (Suc j)) \<notin> rstep R" using rstep_mono[OF R] by force
qed


declare rel_SN_alt.simps[simp del]
declare rel_step.simps[simp del]

lemma rel_SN_imp_SN : assumes "rel_SN (R,S)" shows  "SN (rstep R)"
proof
  fix f
  assume "\<forall> i. (f i, f (Suc i)) \<in> rstep R"
  hence "\<And> i. (f i, f (Suc i)) \<in> rel_step (R, S)" unfolding rel_step.simps by blast  
  thus False using assms unfolding rel_SN_def SN_defs by blast
qed


lemma left_right_rel_step : "(rstep (R \<union> S))^* O rstep R O (rstep (R \<union> S))^* \<subseteq> (rel_step (R,S))^+" (is "?RS^* O ?R O _ \<subseteq> _")
proof(clarify, simp)
  fix x1 x2 x3 x4
  assume x12: "(x1,x2) \<in> (rstep (R \<union> S))^*" and x23: "(x2,x3) \<in> rstep R" and x34: "(x3,x4) \<in> (rstep (R \<union> S))^*"
  let ?S = "(rstep S)^*"
  {
    fix x y z
    assume "(y,z) \<in> (rstep (R \<union> S))^*"
    hence "(x,y) \<in> rel_step (R,S) \<longrightarrow> (x,z) \<in> (rel_step (R,S))^+"
    proof (induct)
      case base
      show ?case by auto
    next
      case (step u z)
      show ?case
      proof
        assume "(x,y) \<in> rel_step (R,S)"
        with step have nearly: "(x,u) \<in> (rel_step (R,S))^+" by simp
        from step(2)[unfolded rstep_union]
        show "(x,z) \<in> (rel_step (R,S))^+"
        proof
          assume "(u,z) \<in> rstep R"
          hence "(u,z) \<in> rel_step (R,S)" unfolding rel_step.simps by auto
          with nearly show ?thesis by auto
        next
          assume uz: "(u,z) \<in> rstep S"
          from nearly[unfolded trancl_unfold_right]
          obtain v where xv: "(x,v) \<in> (rel_step (R,S))^*" and vu: "(v,u) \<in> rel_step (R,S)" by auto
          from vu obtain w where vw: "(v,w) \<in> ?S O rstep R" and wu: "(w,u) \<in> ?S" unfolding rel_step.simps by auto
          from wu uz have wz: "(w,z) \<in> ?S" by auto
          with vw have vz: "(v,z) \<in> rel_step (R,S)" unfolding rel_step.simps by auto
          with xv show ?thesis by auto
        qed
      qed
    qed
  } note steps_right = this
  from x23 have "(x2,x3) \<in> rel_step (R,S)" unfolding rel_step.simps by auto
  from mp[OF steps_right[OF x34] this] have x24: "(x2,x4) \<in> (rel_step (R,S))^+" .
  with x12 show "(x1,x4) \<in> (rel_step (R,S))^+" 
  proof (induct arbitrary: x4, simp)
    case (step y z) 
    from step(2)
    have "(y,x4) \<in> (rel_step (R,S))^+"
      unfolding rstep_union
    proof
      assume "(y,z) \<in> rstep R"
      hence "(y,z) \<in> rel_step (R,S)" unfolding rel_step.simps by auto
      with step(4) show ?thesis by auto
    next
      assume yz: "(y,z) \<in> rstep S"
      from step(4)[unfolded trancl_unfold_left]
      obtain v where zv: "(z,v) \<in> rel_step (R,S)" and vx4: "(v,x4) \<in> (rel_step (R,S))^*" by auto
      from zv obtain w where zw: "(z,w) \<in> ?S" and wv: "(w,v) \<in> rstep R O ?S" unfolding rel_step.simps by auto
      from yz zw have "(y,w) \<in> ?S" by auto
      with wv have "(y,v) \<in> rel_step (R,S)" unfolding rel_step.simps by auto
      with vx4 show ?thesis by auto
    qed
    from step(3)[OF this] show ?case .
  qed
qed 

end